
import io
import sys
from time import sleep
from datetime import datetime
from polzyFunctions.scripts.utils import db, models
from polzybackend.models import GamificationActivity, GamificationUserStats


save_stdout = sys.stdout  # saves original stdout
sys.stdout = io.StringIO()  # using dummy handler for stdout
from config import Config  # This command throws a print statement therefore using dummy stdout handler
sys.stdout = save_stdout  # restoring original stdout handler



def print_args():
    print("""    This program is used to update Gamification Statistics with unprocessed Events Stored in GamificationActivity Table
    To update statistics in a loop with a predifined process sleep time   = python GamificationStatsUpdater.py --sleep <time in seconds>
    To update statistics only for one time = python GamificationStatsUpdater.py --sleep 0""")


def get_activites():
    """
    Returns unprocessed gamification Activities.
    :return:
    """
    activites = []
    processed = 0
    for activity in db.session.query(GamificationActivity).all():
        if not activity.is_processed:
            activites.append(activity)
        else:
            processed += 1
    print(f'Total activites are {str(processed + len(activites))} out of which {str(processed)} are already processed'
          f' and {str(len(activites))} are ready to be processed.')
    return activites


def update_statistics():
    """
    Updates unprocessed statistics
    :param seconds: Time in seconds for sleep before rerun loop
    :return:
    """
    activites = get_activites()
    users = set()  # using set in order to get count of unique users at the end
    done = 0
    total = len(activites)
    for activity in activites:
        user_id = GamificationUserStats.create_or_update_row(activity, commit=False)  # creates or updates user stats
        if not user_id:  # if user_stats has none objects that means event type is to be ignored
            print(f"Activity {activity.id} skipped because of ignored event type.")
            activity.set_processed(commit=False)  # sets activity as processed
            continue
        users.add(user_id)
        activity.set_processed(commit=False)
        done += 1
        if done % 100 == 0:
            print(f"{done}/{total}")
    GamificationUserStats.commit()
    print(f"Total {str(len(activites))} activities updated for {str(len(users))} unique users.")


def run_loop(seconds=0):
    exceptions = 0
    excpetion_threshold = 100
    while True:
        try:
            update_statistics()
            exceptions = 0
        except Exception as ex:
            exceptions += 1
            print(f"Exception {str(exceptions)}: {str(ex)}")
            logger.exception(f"Exception {str(exceptions)}: {str(ex)}")
            if exceptions >= excpetion_threshold:
                print(f'Exiting loop as total continuous exceptions has reached threshold {str(excpetion_threshold)}')
                sys.exit()
        if not seconds:  # if seconds is 0 then we not need to run this program in loop
            break
        else:
            print(f"Sleeping for {str(seconds)} seconds. Statistics database updated at",
                  datetime.now().strftime("%d/%m/%Y %H:%M:%S"))
            sleep(seconds)


if __name__ == "__main__":
    if len(sys.argv) < 2:
        print_args()
        update_statistics()
    elif "sleep" in sys.argv[1] and len(sys.argv) > 2:
        try:
            seconds = int(sys.argv[2])
        except:
            print("Integer value needed for sleep. e.g. python GamificationStatsUpdater.py --sleep 10")
            sys.exit()
        run_loop(seconds)
    else:
        print_args()



import os
import sys
import json
import xlrd
import getopt


def print_args():
    print("""   This program is used to convert 2 columns of excel sheet to json data as key: value pair.
    
    Usage:
    python Excel2Json.py --file <filename and path> [options]
    
    Options:
    --file         <filename>      give excel filename with complete path
    --sheetname    <sheet name>    If you want data for a specific sheet only mention sheet name
    --start-row    <Row Number>    Input the starting Row number from where we need json data. (Row index starts from 1)
    --end-row      <Row Number>    Input the ending Row number till where we need json data. (Row index starts from 1)
    --key-column   <Column Number> Input the Column number of "Key" for json data. (Column index starts from 1)
    --value-column <Column Number> Input the Column number of "Value" for json data. (Column index starts from 1)
    """)


def args_read():
    # reading args
    filename = None  # assigning default value for all necessary parameters
    sheetname = None
    start_row = 1
    end_row = None
    key_column = 1
    value_column = 2
    l_args = sys.argv[1:]

    try:
        opts, args = getopt.getopt(l_args, "", ["file=",
                                                "sheetname=",
                                                "start-row=",
                                                "end-row=",
                                                "key-column=",
                                                "value-column="
                                                ])
    except getopt.GetoptError as err_det:
        print_args()
        print("Error in reading parameters:" + str(err_det))
        sys.exit("Wrong parameters - exiting")
    if opts:
        for opt, arg in opts:
            if opt == "--file":
                if not os.path.exists(arg):
                    print(f"Please input the correct filename with path {arg}")
                    sys.exit()
                filename = arg
            elif opt == "--sheetname":
                sheetname = arg
            elif opt == "--start-row":
                start_row = int(arg)
            elif opt == "--end-row":
                end_row = int(arg)
            elif opt == "--key-column":
                key_column = int(arg)
            elif opt == "--value-column":
                value_column = int(arg)
            else:
                print("Please select correct options.")
                print_args()
                sys.exit()
    if filename:
        excel_to_json(filename, sheetname, key_column, value_column, start_row, end_row)
    else:
        print("Please input filename.")
        print_args()


def sheet_to_json(sheet: xlrd.sheet, sheetname: str, key_column, value_column, start_row, end_row):
    if not end_row:  # getting end row if end row is not supplied
        end_row = sheet.nrows
    key_column -= 1  # subtracting 1 to match the python index with excel index
    value_column -= 1
    start_row -= 1
    js = {}
    for row in range(start_row, end_row):
        try:
            key = int(sheet.cell(row, key_column).value * 100)
            value = int(sheet.cell(row, value_column).value * 100)
            js[key] = value
        except Exception as ex:
            print(f"Exception in row {str(row)}: {str(ex)}")
    with open(sheetname+".json", "w")as file:  # using sheetname as filename
        json.dump(js, file, indent=4)
    print(sheetname+".json created successfully.")


def excel_to_json(filename, sheetname, key_column, value_column, start_row, end_row):
    wb = xlrd.open_workbook(filename)
    shts = wb.sheet_names()
    if sheetname:  # if sheetname is supplied then create json file only for that sheet
        if not sheetname in shts:
            print(f"{sheetname} sheet is not in {filename}")
            sys.exit()
        lsheet = wb.sheet_by_name(sheetname)
        sheet_to_json(lsheet, sheetname, key_column, value_column, start_row, end_row)
    else:
        for sht in shts:
            lsheet = wb.sheet_by_name(sht)
            sheet_to_json(lsheet, sht, key_column, value_column, start_row, end_row)
    print("Done.")


if __name__ == "__main__":
    if len(sys.argv) > 1:
        args_read()
    else:
        print_args()



from flask import jsonify, request, current_app, send_file
from polzybackend.general import bp
from polzybackend.utils.import_utils import all_stages
from polzybackend import auth, models, db
from datetime import datetime
import os
from uuid import uuid4
from polzyFunctions.GlobalConstants import logger

@bp.route('/stages')
def stages():
    #
    # returns list of all available stages for login
    #

    try:
        # get all stages
        stages = all_stages()()
        current_app.logger.debug(f"Value of stages: {stages}")

    except Exception as e:
        current_app.logger.warning(f'Failed to get All Stages: {e}')
        stages = []

    return jsonify(stages), 200


@bp.route('/values', methods=['POST'])
@auth.login_required
def values():
    #
    # returns value list
    #

    # get post data
    data = request.get_json()

    try:
        # get parent instance from app store
        # try policies first
        instance = current_app.config['ANTRAGS'].get(data['instanceId'])
        if instance is None:
            # try antrags then
            instance = current_app.config['POLICIES'].get(data['instanceId'])
            if instance is None:
                raise ValueError(f'Instance with id {data["instanceId"]} not found in PoLZy storage. Most probably app restarted.')

        # get value list
        result = instance.getValueList(data.get('valueListName'))
        return jsonify(result), 200
    except Exception as e:
        current_app.logger.exception(f'Faild to get value-list for paylod {data}\n{e}')
    
    return jsonify({'error': f'Failed to get value-list'}), 400


@bp.route('/upload', methods=['POST'])
@bp.route('/upload/<string:parent_id>/<path:file_type>', methods=['POST'])
@auth.login_required
def upload(parent_id=None, file_type=None):
    # get file
    file = request.files.get('file')
    if file is None:
        return jsonify({'error': 'Request does not contain dataFile'}), 400

    # save file
    try:
        user = auth.current_user()
        filename_parts = (str(uuid4()), file.filename.split('.')[-1])
        path_to_file = os.path.join(current_app.config['UPLOADS'], '.'.join(filename_parts))
        file.save(path_to_file)
        # create file instance in db
        file_db = models.File.new(
            user=auth.current_user(),
            id=filename_parts[0],
            filename=file.filename,
            parent_id=parent_id,
            file_type=file_type,
        )
        return {'OK': f'File {file_db.filename} saved with id {file_db.id}'}, 200
    except Exception as error:
        current_app.logger.error(f'Failed to upload file "{file.filename}" by {user}: {error}.')
        return jsonify({'error': 'File upload failed'}), 400
    

@bp.route('/documents/<string:file_id>', methods=['GET', 'POST', 'DELETE'])
@auth.login_required
def manage_file(file_id):
    # get file record
    file = models.File.query.get(file_id)
    try:
        ext = file.filename.split('.')[-1]
    except AttributeError:
        ext = ""
    path_to_file = os.path.join(current_app.config['UPLOADS'], f'{file_id}.{ext}')

    # edit file type
    if request.method == 'POST':
        payload = request.get_json()
        file.type = payload.get('fileType')
        try:
            db.session.commit()
        except Exception as ex:
            logger.critical(f"Exception while committing changes to db: {ex}")
            db.session.rollback()
        return {}, 200

    # delete file
    if request.method == 'DELETE':
        # delete file record
        try:
            db.session.delete(file)
            try:
                db.session.commit()
            except Exception as ex:
                logger.critical(f"Exception while committing changes to db: {ex}")
                db.session.rollback()
        except UnmappedInstanceError as e:
            current_app.logger.critical(f"File {file} has not mapped instance. Error: {e}")
        return {}, 200

    # get file
    return send_file(
        path_to_file,
        attachment_filename=file.filename,
    )


@bp.route('/remotedocuments', methods=['POST'])
@auth.login_required
def remote_documents():
    #
    # returns link to a remote dokument or a ziped bunch of remote documents  
    #

    # get post data
    data = request.get_json()

    try:
        # get parent instance from app store
        # try policies first
        instance = current_app.config['ANTRAGS'].get(data['parentId'])
        if instance is None:
            # try antrags then
            instance = current_app.config['POLICIES'].get(data['parentId'])
            if instance is None:
                raise ValueError(f'Instance with id {data["parentId"]} not found in PoLZy storage. '
                                 f'Most probably app restarted.')

        # get path
        path_to_document = instance.getRemoteDocuments(data.get('documentsId'))
        return send_file(path_to_document)

    except Exception as e:
        current_app.logger.exception(f'Faild to get remote documents for paylod {data}\n{e}')
    
    return jsonify({'error': f'Faild to get remote documents'}), 400


@bp.route('/generateeml', methods=['POST'])
@auth.login_required
def generate_eml():
    #
    # returns link to a remote dokument or a ziped bunch of remote documents
    #

    # get post data
    data = request.get_json()

    try:
        # get parent instance from app store
        # try policies first
        instance = current_app.config['ANTRAGS'].get(data['parentId'])
        if instance is None:
            # try antrags then
            instance = current_app.config['POLICIES'].get(data['parentId'])
            if instance is None:
                raise ValueError(f'Instance with id {data["parentId"]} not found in PoLZy storage. '
                                 f'Most probably app restarted.')

        # get path
        path_to_document = instance.generateEml(data.get('documentsId'))
        return send_file(path_to_document)

    except Exception as e:
        current_app.logger.exception(f'Faild to get remote documents for paylod {data}\n{e}')

    return jsonify({'error': f'Faild to get remote documents'}), 400


@bp.route('/generateantrageml', methods=['POST'])
@auth.login_required
def generate_antrag_eml():
    # returns link to a eml file
    data = request.get_json()
    try:
        instance = current_app.config['ANTRAGS'].get(data['parentId'])
        if instance is None:
            # try antrags then
            instance = current_app.config['POLICIES'].get(data['parentId'])
            if instance is None:
                raise ValueError(f'Instance with id {data["parentId"]} not found in PoLZy storage. '
                                 f'Most probably app restarted.')

        # get path
        path_to_document = instance.generateAntragEml()
        return send_file(path_to_document)

    except Exception as e:
        current_app.logger.exception(f'Faild to get remote documents for paylod {data}\n{e}')

    return jsonify({'error': f'Faild to get remote documents'}), 400


from dataclasses import dataclass


@dataclass
class GlobalSystemConstants():

    @classmethod
    def getAllStages(self) -> list:
        """
        You need to subclass this class and add stages (e.g. DEV, Pre-Quality, Production).
        :return:
        """
        return []



from polzyFunctions.Dataclasses.Antrag import Antrag
from polzyFunctions.Activities.Activity import Activity
from polzyFunctions.GamificationActivityRecorder import recordActivityDecorator
from polzyFunctions.AntragActivityRecorder import recordAntragDecorator
from polzyFunctions.GlobalConstants import logger



class AntragActivity(Activity):
    """
    All activities are called with an instance of Polizze and effectiveDate (Effective date)

    The first call for each activity is to method 'checkIsActivityValidForPolicy'.
        For instance a cancel-activity will switch itself of, if the policy is already in status cancelled.
        The activity will not be offered to the user.

    Next a remote system would call 'getFieldsForActivity' method to know, which attributes need to be provided
        for this activity to run.

    It is advisable to call 'checkAndUpdateInput' if the attribute of an updated field has the value bool True
        in field inputTriggersComplexUpdates

    With method 'executeActivity' the activity is executed.

    Internally each activity will use or implement additional methods, e.g.
        'addFieldForActivity'
        'createFieldcatalogForActivity'

    """

    def __init__(self, antrag: Antrag, **kwargs):
        super(AntragActivity, self).__init__()
        self.antrag = antrag

    def checkIsActivityValidForStatus(self) -> bool:
        """
        Define, whether this activity is valid for the current status of the current instance. Each instance must
        implement this method
        :return: True if valid/active, False if not valid/activ
        """
        logger.exception(f"Activity not properly implemented! Doesn't know if it's valid for status or not! "
                         f"{self.__class__}")
        raise NotImplementedError((f"Activity not properly implemented! Doesn't know if it's valid for status or not! "
                                   f"{self.__class__}"))

    def setSingleFieldValue(self, fieldNameToSearchFor, valueToSet):
        """
        When we get new values from the frontend or from within the acivities logic, they are processed here

        :param fieldNameToSearchFor:
        :param valueToSet:
        :return:
        """
        self.updateFieldValues({fieldNameToSearchFor: valueToSet})

    def setSingleFieldFieldType(self, fieldNameToSearchFor, fieldVisibilityTypeToSet):
        """
        Sets visibility of a field
        :param fieldNameToSearchFor:
        :param fieldVisibilityTypeToSet: Any of the values of FieldVisibilityTypes (e.g. hidden, visible)
        :return:
        """
        lFeld = self.antrag.Fields.getField(name=fieldNameToSearchFor)
        if lFeld:
            lFeld.fieldVisibilityType = fieldVisibilityTypeToSet
        else:
            logger.warning(f"Should have set FieldType {fieldVisibilityTypeToSet} for field {fieldNameToSearchFor}. But field"
                           f"is not there")

    @recordAntragDecorator
    @recordActivityDecorator
    def executeActivity(self) -> bool:
        """
        Executes this activity. See description in superclass.
        :return: True if successful, False if not successful.
        """
        return super().executeActivity()


from enum import Enum

class CommonFieldnames(Enum):
    """
    Common Fieldnames, that are used in many products (it not all).

    """
    policyBeginDate = "Versicherungsbeginn"
    policyEndDate = "Versicherungsende"
    mainDueDate = "Hauptfälligkeit"
    businessCaseType = "GeschäftsfallArt"
    previousPolicyNumber = "Vorpolizzennummer"
    premium = "premium"
    paymentFrequency = "Zahlungsfrequenz"
    expertMode = "Erweitern"    # "ExpertenModus"
    commissionAccount = "ProvKto"

    ## Person fields:
    Partner = "Partner"
    partnerName = "partnerName"
    partnerNumber = "partnerNumber"
    birthDate = "birthDate"
    gender = "gender"
    lastName = "lastName"
    firstName = "firstName"

    ## Company fields
    companyName = "companyName"
    companyType = "companyType"
    registrationNumber = "registrationNumber"

    ## Address fields:
    addressDict = "addressDict"
    addressID = "addressID"           # official addressID in Backend-System
    addressNumber = "addressNumber"   # Polzy-Internal, temporary addressNumber
    country = "country"
    postCode = "postCode"
    city = "city"
    street = "street"
    streetNumber = "streetNumber"
    houseNumber = "houseNumber"
    telefon = "telefon"
    email = "email"

    # Fields for "Senden-An-Bestand"
    underWriterWarnings = "AnnahmeTexte"
    policyNumber = "policyNumber"
    applicationNumber = "applicationNumber"
    pac = "pac"


from threading import Thread
from time import sleep
import configparser
import logging
from pathlib import Path
from os import getcwd


class Singleton(type):
    _instances = {}

    def __call__(cls, *args, **kwargs):
        if cls not in cls._instances:
            cls._instances[cls] = super(Singleton, cls).__call__(*args, **kwargs)
        return cls._instances[cls]


class LogLevelUpdater(metaclass=Singleton):
    def __init__(self):
        self.loggers = []
        self.threads = []
        self.create_thread = False
        self.log_levels = {'info': logging.INFO, 'debug': logging.DEBUG, "error": logging.ERROR,
                           'critical': logging.CRITICAL, 'warning': logging.WARNING}
        self.config_file = str(self.findConfigFileInPath())
        self.config = configparser.ConfigParser()
        self.config.read(self.config_file)
        self.filelevel = self.config["Default"].get("filelevel", '').lower().strip() or 'debug'
        self.channellevel = self.config["Default"].get("channellevel", '').lower().strip() or 'info'
        self.saveJson = None
        self.getSaveJson()

    @staticmethod
    def findConfigFileInPath():
        lConfigFilename = "run_setting.ini"
        lConfigPath = Path(getcwd()).joinpath(lConfigFilename)
        if lConfigPath.exists():
            return lConfigPath

        lConfigPath = Path(getcwd()).parent.joinpath(lConfigFilename)
        if lConfigPath.exists():
            return lConfigPath

        print(f"Error: Can't locate run_setting.ini in {getcwd()} or parent folder. Please check working directory of "
              f"this python instance. should be /fasifu/ or /fasifu/tests.")

    def updateLogLevel(self):
        for logger in self.loggers:
            for handler in logger.handlers:
                if isinstance(handler, logging.FileHandler):
                    handler.setLevel(self.log_levels[self.filelevel])
                elif isinstance(handler, logging.StreamHandler):
                    handler.setLevel(self.log_levels[self.channellevel])
            logger.propogate = False
        print("File log level:", self.filelevel)
        print("Channel log level:", self.channellevel)

    def checkAndUpdateLogLevel(self):
        while True:
            sleep(60)
            self.config.read(self.config_file)
            new_filelevel = self.config["Default"].get("filelevel", '').lower().strip() or 'info'
            new_channellevel = self.config["Default"].get("channellevel", '').lower().strip() or 'info'
            if not new_channellevel == self.channellevel and not new_filelevel == self.filelevel:
                self.channellevel = new_channellevel
                self.filelevel = new_filelevel
                self.updateLogLevel()

            self.getSaveJson()

    def getSaveJson(self):
        # Reading and updating json flag
        self.saveJson = self.config["Default"].get("saveJson") or False
        if self.saveJson.lower().strip() == 'true':
            self.saveJson = True
        else:
            self.saveJson = False

    def startThread(self):
        self.updateLogLevel()
        if not self.threads and self.create_thread:
            t = Thread(target=self.checkAndUpdateLogLevel)
            t.daemon = True
            t.start()
            self.threads.append(t)



import json
import uuid
import locale
from dataclasses import dataclass
from polzybackend import messenger
from polzybackend.models import User, AntragActivityRecords
from datetime import datetime, timedelta
from polzyFunctions.utils.counter import Counter
from polzyFunctions.Dataclasses.PersonRoles import Roles
from polzyFunctions.Activities.Activity import Activity
from polzyFunctions.GlobalConstants import GlobalConstants, logger
from polzyFunctions.Dataclasses.AntragsStatus import AntragsStatus
from polzyFunctions.AntragActivityRecorder import recordAntragDecorator
from polzyFunctions.Dataclasses.CommonFieldnames import CommonFieldnames
from polzyFunctions.GamificationActivityRecorder import recordActivityDecorator
from polzyFunctions.Activities.ActivitiesDataClasses import InputFields, FieldDataType, FieldVisibilityTypes, \
    InputFieldTypes, FieldDefinition


@dataclass
class Antrag():
    """
    THE main component of LeZyTOR (Tariff Online Calculator). In this class (and all inheriting classes) we handle
    instances of fastOffers.
    Main attributes:
    Fields: Fieldcatalog - please see documentation in there
    Activities: Activities for this instance. Usually you'd have a few static activities and most of the activities
           would be depending on a) the product or b) the status of the instance. In this list of Activities are only
           activities, which are currently allowed for this instance.
    """
    id: uuid.UUID          # Unique ID
    user: User             # User from Fronteend
    productName: str       #
    status: str            # current status of this instance.
    # Key to use in configurationProvider to search for configuration values. Usually you'd have one configuration
    # provider per insurance company as fallback and e.g. one configuration provider for each sales channel.
    # Depends largely on the use-case of POLZY
    configurationProviderKey: str
    Fields: InputFields()    # Field catalog for this product
    Activities: [Activity]   # Currently allowed activities in this instance.
    # The search string contains a concatenation (with space) of all relevant properties for a quick search when we
    # want to reload persisted instances from the database.
    searchstring: str = ""

    def __init__(self, user):
        self.user = user

        self.status = ""  # take a look sir
        self.Activities = []  # take a look sir
        # bufferedFieldValuesForComparisonWithFrontend is updated after each roundtrip with the frontend. When we
        # get a request to send an instance via parseToFrontend-Method we store the latest values in this dict.
        # Upon receiving the new values from the frontend in updateFieldValues we can use this dict to identify or
        # compare changed values.
        self.bufferedFieldValuesForComparisonWithFrontend = {}
        self.Fields = InputFields()

        self.id = uuid.uuid4()
        # The Tag is used in frontend (and saved to database) to set custom tags in order to find the instance
        # later, independently of the product, id or other search fields. An example of usage would be a clerk, who
        # calculates multiple offers and multiple Products for Mrs. Smith (not yet a customer). He would tag them with
        # "Smith" and later use search functionality to find all of them.
        self.tag = None
        # Additionally to the unique id (self.id) we generate a human readable instance number. Can be overwritten.
        self.antragsnummer = Counter().get_number()
        # In this dict we collect warning for underwriters, e.g. when a person has a negative mark or certain
        # field combinations raise warning. When those warnings are set the instance wouldn't automatically be sent
        # to Backend system but would raise those warnings in a final screen. Apart from the dict itself there is no
        # logic inside PoLZy. You'll have to develop the reaction to this field in your installation.
        self.underWriterWarnings = {}
        self.latestDBTimestamp = None

    def loadActivitiesFromDict(self, activities: dict):
        """
        This method is used in persistence function. It takes a dictionary as input which contains classname as key &
        its attributes as value. Classname is used to create/get class instance and then value is supplied to that
        class instance in order to load its attributes.

        This method can be overwritten in any subclass if you want to execute several tasks after all fields and
        all activties have been loaded
        :param activities: Dict of Activities loaded from DB-Field of the antrag-Instance as JSON.
        :return:
        """
        for classname, value in activities.items():
            instance = self.get_activity(classname, optional=True)
            if instance:
                # if instance of same activity is already their than update it
                instance = instance[0]
                logger.debug(f"Found existing instance of {classname}. Using it to update persistence values.")
                instance.loadFromJsonFromPersistence(value)
            else:  # if no existing instance of activity found than create a new and add to list
                logger.debug(f"No instance found of {classname}. Creating New.")
                class_ = "fasifu.Activities." + classname  # used to dynamically import module
                if classname == "AntragClone":  # AntragClone class's module name is not same. Hence updating it
                    class_ += "Copy"
                mod = __import__(class_, fromlist=[''])  # Importing class object from module
                instance = getattr(mod, classname)(self)  # creating Activity instance
                instance.loadFromJsonFromPersistence(value)
                self.Activities.append(instance)
            logger.debug(f"{classname} loading successful.")

    def setSearchString(self):
        """
        Method to set the search string, e.g. during initializiation or when the Tag was changed. Implement in your
        own installation. Most probably specific for each product.
        :return:
        """
        return""

    def handleUnderWriterWarnings(self, section, warningToAdd, textToAdd=None):
        """
        This dict will hold warnings for underwriters. It has a section, within the section a key and a text.
        :param section: A freely usable term, that will help the underwriter to understand, where the problem happened
        :param warningToAdd: a technical key (used to prevent continuous nagging of the user with the same message
        :param textToAdd: The text, that will be displayed for the underwriters
        :return: true if a change happend (remove or add). False if there was no change to the dict.
        """

        if not self.underWriterWarnings.get(section):
            self.underWriterWarnings[section] = {}

        if not textToAdd:  # We shall remove this warning, if exists
            try:
                del self.underWriterWarnings[section][warningToAdd]
                if not self.underWriterWarnings.get(section, "empty") != "empty":
                    del self.underWriterWarnings[section]
                return True   # There was a key, which is now removed.
            except KeyError:
                pass
            return False   # The key did not exist, so no changes happened.

        if self.underWriterWarnings[section].get(warningToAdd) == textToAdd:
            return False
        self.underWriterWarnings[section][warningToAdd] = textToAdd
        return True

    @property
    def uuid(self):
        return str(self.id)

    def set_user(self, user):
        self.user = user

    @recordAntragDecorator
    def setCustomTag(self, tag):
        logger.debug(f"Tag is set to {tag}")
        self.tag = tag

    def updateAfterClone(self):
        """
        After an instance was cloned do here some cleanup work. Can and should be overwritten in each installation.
        :return:
        """

        # generate new ID
        self.id = uuid.uuid4()
        # clear tag
        self.tag = None

        # Copied Antrag needs new Antragsnummer
        self.antragsnummer = Counter().get_number()
        self.productName = self._generateProduktNameWithAntragsnummer()
        self.setSearchString()

    def getValueList(self, listName):
        """
        long lists should not be transferred to the frontend in each request. Instead we can hand them over once. Each
        listName is unique. You should provide here the values for the frontend.
        :param listName: name of the list-object
        :return: Values of the list object for the frontend
        """
        # should be defined within specific antrag class
        raise ValueError(f'List with name {listName} not defined')

    def getClauses(self) -> list:
        """
        Method (so far only used for printing) to retrieve a list of clauses and their texts.
        :return: List of lists [["1000k", "Special conditions"], ["1001k", "very special conditions]
        """
        raise NotImplementedError(f"The class {self.__class__.__name__} does not provide implementation for getClauses")

    def getPremiumValues(self) -> dict:
        """
        Method (so far only for printing) to retrieve the annual premium, taxes and
        premium per payment frequency (if not annual)
        :return: as dict {"annualPrem": 13172,72, "taxPercent", 4.12, "taxValue": 512.23}
                 plus optional "premiumPerPeriod": 1097,67 if payment frequency is not annual.
        """
        raise NotImplementedError(f"The class {self.__class__.__name__} does not provide implementation for getClauses")

    def getRemoteDocuments(self, documents_id: list):
        # This method must be implemented in subclass
        # This method should be used to get/generate documents and return its full path.
        # Later it will be used for downloading function
        return

    # POLZY - empty with documentation
    def generateEml(self, documents):
        # This method must be implemented in subclass
        # This method should be used to generate eml file with documents as attachment and return full path of the file.
        # Later it will be used for downloading function of eml & that eml can be used to directly forward mail
        return

    # POLZY - empty with documentation.
    def generateAntragEml(self):
        # This method must be implemented in subclass
        # This method should be used to generate eml file with link to antrag and return full path of that eml file.
        # Later it will be used for downloading function of eml & that eml can be used to directly forward mail
        return

    def get_activity(self, activity_name, optional=False):
        """
        Retrieve an activity from the list of Activities in this instance by class name
        :param activity_name: class name of the desired activity
        :param optional: Don't raise an error if the activity can't be found
        :return: the activity
        """
        activityObjects = [x for x in self.Activities if x.__class__.__name__ == activity_name]
        if not activityObjects:
            activityObjects = [x for x in self.Activities if x.name == activity_name]
            if not activityObjects and not optional:
                logger.critical(f"Cannot find activity: {activity_name}", stack_info=True)
        return activityObjects

    def getPreviousVersions(self):
        """
        gets list of all AntragActivityRecords instances for current antragsnummer
        :return:
        """
        fmt = "%d-%m-%Y %H:%M:%S"
        versionsDict = {
            record.id: [
                record.timestamp.strftime(fmt),
                User.get_user_by_id(record.user_id).get("name"),
                record.status]
                    for record in AntragActivityRecords.getAllVersions(self.antragsnummer)
        }
        return versionsDict

    def formatToString(self):
        """
        String-Representation (used in __repr__ and __str__) about this instance.
        :return:
        """
        return f"{self.status} {self.antragsnummer} {self.lineOfBusiness}"

    def __repr__(self):
        return self.formatToString()

    def __str__(self):
        return self.formatToString()

    # POLZY
    def getSpeedometerValue(self):
        """
        Derives speedometer value

        Possible returned values:
        None -- do not show speedometer in front-end
        0-1000 -- show speedometer with given value

        Should be implemented by each Antrag class. Default = None
        :return:
        """

        return None

    def parseToFrontend(self):
        """
        Creates the JSON that is sent as payload to the frontend
        :return: the JSON
        """

        try:
            lActivitiesJson = [a.toJSON() for a in self.Activities]
            logger.debug(f"lActivitiesJSON hat den Wert: {lActivitiesJson}")
        except Exception as e:
            lActivitiesJson = {}
            logger.critical(f"Fehler bei Aktiviäten JSONisierung: {e}")

        frontendDict = {
            'id': self.uuid,
            'tag': self.tag,
            "number": self.antragsnummer,
            'status': self.status,
            'product_line': {
                'name': self.lineOfBusiness,
                'attributes': {
                    'Produkt': self.productName
                },
            },
            'possible_activities': lActivitiesJson,
            'fields': self.parseToFrontendFieldsWithoutGroup(),
            'speedometerValue': self.getSpeedometerValue(),
        }

        lFieldGroups = self.parseToFrontendFieldGroups()
        if lFieldGroups:
            frontendDict.update(lFieldGroups)
            frontendDict.update(self.parseToFrontendFieldGroupFields())

        logger.debug(f"Antrag for Frontend class {self.__class__.__name__} is:\n {frontendDict}")

        return frontendDict

    def parseToFrontendFieldsWithoutGroup(self) -> list:
        """
        Fields, that are not part of field groups and not hidden are jsonified here
        :return:
        """

        # get list of fields which are:
        # 1) not groups
        # 2) not in a group
        # 3) not hidden

        lList = list(map(
            lambda feld: feld.toJSON(),
            filter(
                lambda feld: not feld.isGroupField and not feld.group and feld.fieldVisibilityType != FieldVisibilityTypes.hidden,
                self.Fields.getAllInputFields()
            )
        ))

        return lList

    def parseToFrontendFieldGroups(self) -> dict:
        """
        Loops through the list of Antrags fields and identifies group-fields (if any).
        If they exist, a list of group-fields is returned, where empty field-groups (with no displayable fields) are
        ommitted.
        :return:
        """
        lList = self.Fields.getAllFieldGroups()
        lAllInputFields = self.Fields.getAllInputFields()

        lDropFieldGroupEntries = []
        for fieldGroup in lList:
            # Field-Groups, that are not displayed anyway should not be transferred:
            if not fieldGroup.value and fieldGroup.name != CommonFieldnames.expertMode.value and fieldGroup.name != CommonFieldnames.AussagenGroup.value:
                lDropFieldGroupEntries.append(fieldGroup)
                continue

            # Field-Groups, that don't have any visible fields inside them shall not be transferred:
            fieldList = [x.name for x in lAllInputFields if x.group == fieldGroup.name
                         and x.fieldVisibilityType != FieldVisibilityTypes.hidden]
            if not fieldList and not fieldGroup.name == CommonFieldnames.expertMode.value and not fieldGroup.name == CommonFieldnames.AussagenGroup.value:
                lDropFieldGroupEntries.append(fieldGroup)

        for lDropfieldGroupEntry in lDropFieldGroupEntries:
            lList.remove(lDropfieldGroupEntry)

        if lList:
            return {"field_groups": [x.toJSON() for x in lList]}

        return {}

    def parseToFrontendFieldGroupFields(self) -> dict:
        """
        Will create one dict for each field-group in the field-catalog
        lDict[<groupname>] = [list of FieldDefinition.toJSON()

        :return:
        """
        lGroups = self.Fields.getAllFieldGroups()
        lReturnDict = {}
        for feld in self.Fields.getAllInputFields():
            if feld.group:
                # If the fieldGroup is not displayed, we shall not send any fields:
                lGroup = [x for x in lGroups if x.name == feld.group][0]
                if not lGroup.value and lGroup.name != CommonFieldnames.expertMode.value and lGroup.name != CommonFieldnames.AussagenGroup.value:
                    continue
                # Don't send fields, which are not displayed on Frontend.
                if feld.fieldVisibilityType != FieldVisibilityTypes.hidden:
                    # add field group to the response
                    if lReturnDict.get(feld.group) is None:
                        lReturnDict[feld.group] = []
                    lReturnDict[feld.group].append(feld.toJSON())

        return lReturnDict

    def getCountOfPotentialPartners(self) -> int:
        """
        This method is used from AntragPartnerDefintionSingle in order to know, how many partners should be initially
        displayed on the screen. Each Antrag may and should override this method.
        :return: Number of Partners, that should be initially displayed in the partner activity cards
        """
        return 1

    def getPartnerRoles(self) -> list:
        """
        List of Partner Roles within this Antrag. Should be overwritten by each Antrag, if needed
        :return: Standard list: [VN, PZ]
        """
        return [Roles.VN, Roles.PZ]

    def updateFieldValuesFromDatabase(self, updateValues):
        """
        This method is called after we load values from database. It's a bit tricky, we must load all the values but
        shall not execute any derviations (e.g. AnzahlBetriebsleiter from DU-Variant)
        :param updateValues: the field list from database
        :return:
        """
        for k,v in updateValues.items():
            self.bufferedFieldValuesForComparisonWithFrontend[k] = v

        self.updateFieldValues(updateValues)

    def updateFieldValues(self, updateValues):
        """
        Frontend (or anything) sends new values for fields of antrag instance.
        If the values are really new/changed, we execute follow up processes.
        :param updateValues: dict of {<field>:<value>}
        :return: Nothing
        """
        logger.info(f'New Values (most probably from Frontend):\n{updateValues}')

        if not self.status:  # return if status is empty  # take a look sir
            return

        if AntragsStatus.getStatusNumberFromText(self.status) >= AntragsStatus.getStatusNumberFromText(
                AntragsStatus.s800_sent):
            self.announceMessage("Antrag bereits gesendet. Leider keine Änderug mehr möglich.")
            return

        self._updateFieldValues(updateValues)

    def updateSapAttributesFromDatabase(self, *args, **kwargs):  # take a look sir
        return

    def _updateFieldValues(self, updateValues):
        """
        Internal Method to update field values. May be overwritten in each implementation.
        :param updateValues:
        :return:
        """
        for name, value in updateValues.items():
            self._updateSingleFieldValueFromFrontendInternal(name, value)

    def _updateSingleFieldValueFromFrontendInternal(self, name, value):
        """
        This method is called for a single field/value-combination.
        :param name: the fieldname, that was updated
        :param value: the new value
        :return:
        """

        lUpdatedField = self.Fields.getField(name=name)
        self._updateTechnicalFieldValuesAfterChange(lUpdatedField)

    @recordActivityDecorator
    def createFieldcatalogForAntrag(self):
        """
        When called, this method will create the field catalog for this instance.
        :return:
        """
        pass

    def fillCurrentlyPossibleActivities(self):
        """
        Depending on status of this instance, product, client, etc. we'll derive the currently possible activities
        and store them in self.Activities for further use
        :return:
        """
        from polzyFunctions.AntragActivitiesDerive import AntragActivitiesDerive
        lActivities = AntragActivitiesDerive(antrag=self)
        self.Activities = lActivities.getActivitiesForAntrag()

    def _addFieldForActivity(self, field: FieldDefinition):
        """
        This method is called during building the field catalog and will insert one single new field into self.Fields.
        :param field:
        :return:
        """
        self.Fields.addField(field)

    def _updateTechnicalFieldValuesAfterChange(self, field: FieldDefinition):
        """
        When we recive values from the frontend they are in JSON and strings. Internally we know from the field-catalog
        which field types those fields have. Here we convert them into the internal format.
        :param field:
        :return:
        """
        if field.fieldDataType == FieldDataType(InputFieldTypes.TYPEBOOLEAN) or \
                field.fieldDataType == FieldDataType(InputFieldTypes.TYPEFLAGFULLLINE):
            self._handleBooleanField(field)
        elif field.fieldDataType == FieldDataType(InputFieldTypes.TYPENUMERIC):
            self._handleNumericField(field)
        elif isinstance(field.value, datetime):
            field.valueOutput = field.value.strftime(GlobalConstants.dateFormat)
            field.valueTech = field.value

        else:
            field.valueTech = field.value
            field.valueOutput = field.value

    def _handleBooleanField(self, field):
        """
        If the planned field data type is boolean we shall interpret string field values and try to identify
        true or false.
        :param field:
        :return:
        """
        # Wenn der geplante Felddatentyp Boolen ist, aber ein String daher kommt, dann den String umfummeln.
        # TYPEFLAGFULLINE is also boolean, but covers a full line in UI.
        if isinstance(field.value, str):
            # Wenn statt boolean der String daher kommt, dann den string umfummeln.
            if field.value.upper() in ["TRUE", "JA", "ENTHALTEN"]:
                field.value = True
            elif field.value.upper() in ["FALSE", "NEIN", "NONE"]:
                field.value = False
            else:
                logger.critical(f"Value for boolean field {field.name} was {field.value}. "
                                f"Wasn't able to determine True/False from that. I'll assume 'No'.")
                field.value = False

        field.valueTech = field.value
        field.valueOutput = field.value

    def _determineProductName(self, returnValue=None):
        """
        Method to overwrite the productName. Should be subclassed/overwritten in each installation
        :param returnValue: Additional value that can be passed to influence the result.
        :return:
        """
        pass

    @staticmethod
    def _handleNumericField(field):
        """
        Handle potential string input (from frontend) and transform into internal values (Integer, Float) if possible
        :param field:
        :return:
        """
        if not field.value:
            field.value = None
            return

        if field.decimalPlaces > 0 and field.fieldVisibilityType == 2:
            # Output-Field should have Value formatted for the UI
            if isinstance(field.value, float):
                field.value = locale.currency(field.value,
                                              grouping=True, symbol=False)
            field.valueTech = field.value
            field.valueOutput = field.value

        elif field.decimalPlaces > 0:
            Antrag.__handleNumericFieldWithDecimals(field)
        else:
            try:
                field.value = int(field.value)
            except ValueError as e:
                if field.value == 'None':
                    field.value = None
                    field.valueTech = None
                    field.valueOutput = None
                    return
                logger.warning(f"Value {field.value} can't be converted to int. Error was {e}")
            field.valueTech = field.value

            if field.value > 10_000:
                # This is pretty sure a sum insured and should be written as decimal number.
                field.valueOutput = locale.currency(field.value,
                                                    grouping=True, symbol=False)
            else:
                field.valueOutput = locale.format_string(f="%.0f",
                                                         val=field.value,
                                                         grouping=True)

    @staticmethod
    def __handleNumericFieldWithDecimals(field):
        try:
            field.value = float(field.value)
            field.valueTech = field.value
            field.valueOutput = locale.currency(field.value,
                                                grouping=True, symbol=False)
        except ValueError:
            field.value = None
            field.valueTech = None
            field.valueOutput = None

    def deriveDefaultsFromUserOrCompany(self):
        """
        In case the installation has user or company dependent attributes for Antrags-instances here is the
        place to derive them.
        :return:
        """
        pass

    def _toggleFieldsOnOff(self, listOfFieldnamesToToggle: list, OnOffAsBoolean: bool):
        """
        Provide a list of fieldnames and "True" if fields shoulds be on (FieldType = 1) or
        "False" if fields should be off (FieldType = 3
        :param listOfFieldnamesToToggle: List for field name
        :param OnOffAsBoolean: True = On (show), False = Off (hide)
        :return:
        """

        for field in listOfFieldnamesToToggle:
            if OnOffAsBoolean:
                self.setFieldVisible(field)
            else:
                self.setFieldInvisible(field)

    def _toggleFieldGroupOnOff(self, listOfFieldgroupNamesToToggle: list, OnOffAsBoolean: bool):
        """
        Toggle a list of fieldgroups (=Cards on the UI) on or off by setting the value of the field-group to True/False
        :param listOfFieldgroupNamesToToggle: List of the names of the fieldgroups to be effected
        :param OnOffAsBoolean: True = On, False = Off
        :return:
        """
        for fielgroup in listOfFieldgroupNamesToToggle:
            try:
                self.Fields.getField(name=fielgroup).value = OnOffAsBoolean
            except AttributeError:
                logger.critical(f"Field {fielgroup} not found. Typo? Forgot to add .value?")

    def setFieldValue(self, fieldName, newValue, optional=False):
        """
        Write a new value into a field of this instance.

        :param fieldName:
        :param newValue:
        :param optional: don't raise an error/write to log, if field does not exist
        :return:
        """
        try:
            lFeld = self.Fields.getField(name=fieldName)
            lFeld.value = newValue
        except Exception as e:
            if not optional:
                logger.critical(f"Should have written {newValue} into field {fieldName} "
                                f"and failed with error {e}. Most probably typo in Program.")
                raise AttributeError(f"Should have written {newValue} into field {fieldName} "
                                    f"and failed with error {e}. Most probably typo in Program.")
            else:
                logger.debug(f"Should have written {newValue} into field {fieldName} "
                             f"and failed with error {e}. Most probably typo in Program. Optional was set, so Ok!")

    def getFieldValue(self, fieldName, optional=False):
        """
        Retrieve the field.value from fieldName
        :param fieldName:
        :param optional: Don't write log entry, if the field does not exist
        :return:
        """
        try:
            return self.Fields.getField(name=fieldName).value
        except AttributeError:
            if not optional:
                lFields = ", ".join(sorted([x.name for x in self.Fields.getAllInputFields()]))
                logger.exception(f"Called for fieldname = {fieldName}. Doesn't exist! Existing Field names: {lFields}",
                                 stack_info=True)

    def setFieldVisible(self, fieldName):
        """

        :param fieldName:
        :return:
        """
        try:
            lField = self.Fields.getField(name=fieldName)
            lField.fieldVisibilityType = FieldVisibilityTypes.visible
        except AttributeError:
            logger.exception(f"Field {fieldName} should be set visible. Doesn't exist. Most probaly a typo!")

    def setFieldInvisible(self, fieldName, optional=False):
        """

        :param fieldName:
        :param optional: Don't write log, if field doesn't exist in the field catalog
        :return:
        """
        try:
            lField = self.Fields.getField(name=fieldName)
            lField.fieldVisibilityType = FieldVisibilityTypes.hidden
            if lField.value and lField.fieldDataType == FieldDataType(InputFieldTypes.TYPEBOOLEAN):
                lField.value = False
        except Exception:
            if optional:
                logger.debug(f"Field {fieldName} should be set invisible. Doesn't exist. "
                             f"'Optional' is set, so most likely Ok")
            else:
                logger.exception(f"Field {fieldName} should be set invisible. Doesn't exist. Most probaly typo!")

    @staticmethod
    def announceMessage(message, duration=3000, level='default', horizontal='left', vertical='bottom'):
        """
        Will send a toast to the frontend
        :param message: The message text.
        :param duration: None = user needs to click "close" in order to disappear, any other value in MilliSeconds
                         until the Toast disappears
        :param level: 'info' will provide blue background with white font,
                      'default' will provide black background with white font,
                      'success' will bring green backgroudn
                      'error' --> red background
                      'warning' --> orange background
        :param horizontal: "left", "right", "center"
        :param vertical: "bottom", "top"
        :return:
        """
        logger.debug(f"Announce: {message}")
        msg = json.dumps({
            'text': message,
            'autoHideDuration': duration,
            'variant': level,
            'anchorOrigin': {
                'horizontal': horizontal,
                'vertical': vertical,
            }
        })
        messenger.announce(f"data: {msg}\n\n")



from polzyFunctions.GlobalConstants import logger

class LoginActivity:
    # this class is made just to use recordActivityDecorator in order to record login activity
    def __init__(self, user):
        self.user = user
        self.id = user.id
        self.event_details = user.to_json()

    def record_login_activity(self):
        # we are importing direct methods here as importing decorator on starting was causing import error
        try:
            from polzyFunctions.GamificationActivityRecorder import WriteActivity
            writer = WriteActivity(self)
            writer.activityWriter()
            return True
        except Exception as ex:
            logger.debug(f"Gamification Activity can't be written. "
                         f"Most probably no gamification settings provided: {ex}")



import os
from pathlib import Path

#
# configuration settings
#


class Config(object):

    # DB
    SQLALCHEMY_DATABASE_URI = 'sqlite:///' + os.path.join(os.path.abspath(os.getcwd()), 'polzy.db')
    
    # PATHS
    PDF = str(Path(os.path.abspath(__file__)).parent.joinpath("pdfoutput"))
    UPLOADS = os.path.join(os.path.abspath(os.path.dirname(__file__)), 'BatchFiles')
    MEDIA = os.path.join(os.path.abspath(os.path.dirname(__file__)), 'media')
    LOGO = os.path.join(MEDIA, 'logo')
    BADGES = os.path.join(MEDIA, 'badges')
    IMAGES = os.path.join(MEDIA, 'images')

    # Local URI
    HOST = 'http://localhost:5000'
    DOWNLOADS = 'files'
    LOGO_URI = 'logo'

    # flask monitoring dashboard
    DASHBOARD_CONFIG = os.path.join(os.path.abspath(os.path.dirname(__file__)), 'dashboard.cfg')
    DASHBOARD_DATABASE_URI = 'sqlite:///' + os.path.join(os.path.abspath(os.path.dirname(__file__)), 'dashboard.db')
    DEBUG = True



from dataclasses import dataclass


@dataclass()
class Role:
    role: str
    roleName: str


@dataclass()
class Roles:
    """
    Role short texts and long texts
    """
    # Versicherungsnehmer, Policyholder. Usually only one, but may be multiple persons or companies
    VN = Role("VN", "Policy holder")

    # Prämienzahler, Premium Payer
    PZ = Role("PZ", "Premium payer")

    # Versicherte Person, Insured Person. Depending on the product and tarif variant may be mutiple persons.
    VP = Role("VP", "Insured person")

    # Betriebsleiter, Head of department (only valid in BUFT) and not a real role in the backend, more
    # like a special kind of role "VP".
    BL = Role("BL", "Operations manager")

    # Beauftragter, Assignee (various forms of people who act in place of the VN)
    BA = Role("BA", "Representative")


import sys
import getopt
from polzyFunctions.scripts.utils import db, models
from sqlalchemy import func, or_


def print_args():
    print("""    This program is used to add new Notification for specified users & companies in ToastNotification Table
    To update statistics for all companies and users = python AddNotification.py --message <Notification Text>
    Options:
    --message       "<Notification Text>"     Enter message text here. This option is compulsory.

    --company-id    "<id1>, <id2>, <id3>,..." Enter company ids separated by commas here. 
                                              If not supplied notification will be for all companies.
    --user-id       "<id1>, <id2>, <id3>,..." Enter user ids separated by commas here. 
                                              If not supplied notification will be for all users.
    --company-name  "<name1>, <name2>,..."    Enter company name separated by commas here. 
                                              If not supplied notification will be for all companies.
    --user-email    "<email1>, <email2>,..."  Enter user email separated by commas here. 
                                              If not supplied notification will be for all users.
    --type          <badge/default/info...>   Specify toast type badge or default
    
    --duration      <milliseconds>            Enter value in milliseconds. e.g. For 1 second use 1000
                                              Default = 3000
    --horizontal    <left/right>              Horizontal alignment of toast notification. (if type=badge then not valid)
                                              Default = left
    --vertical      <top/bottom>              Vertical alignment of toast notification. (if type=badge then not valid)
                                              Default = top
    
    Note: Make sure to input values of options inside (") double quotes to ignore any error.
    """)


def args_read(l_search_parameter):
    l_args = sys.argv[1:]

    try:
        opts, args = getopt.getopt(l_args, "", ["message=",
                                                "company-id=",
                                                "user-id=",
                                                "company-name=",
                                                "user-email=",
                                                "type=",
                                                "duration=",
                                                "horizontal=",
                                                "vertical="
                                                ])
    except getopt.GetoptError as err_det:
        print("Error in reading parameters:" + str(err_det))
        print_args()
        sys.exit("Wrong parameters - exiting")
    if opts:
        for opt, arg in opts:
            if l_search_parameter == opt:
                return arg
            if "--" + l_search_parameter == opt:
                if arg:
                    return arg
                else:
                    return None
    return None


def sanitize_ids(text):
    if text:
        return [id_.strip() for id_ in str(text).split(",")]
    return None


def get_message():
    message = args_read("message")
    if not message:
        print("\nPlease supply notification text with option --message\n")
        print_args()
        sys.exit()
    return message


def get_company():
    companies = sanitize_ids(args_read("company-id"))
    if not companies:
        companies = []
        names = sanitize_ids(args_read("company-name")) or []
        for name in names:
            company = db.session.query(models.Company).filter(func.lower(models.Company.name) == func.lower(name)).first()
            if company:
                companies.append(company.id)
            else:
                print(f"No company found with name {name}")
    return companies


def get_user():
    users = sanitize_ids(args_read("user-id"))
    if not users:
        users = []
        emails = sanitize_ids(args_read("user-email")) or []
        for email in emails:
            user = db.session.query(models.User).filter(func.lower(models.User.email) == func.lower(email)).first()
            if user:
                users.append(user.id)
            else:
                print(f"No user found with email: {email}")
    return users


def update_notification(message, type, duration, horizontal, vertical, companies=None, users=None):
    models.ToastNotifications.new(message, type=type, duration=duration, horizontal=horizontal,
                                    vertical=vertical, company_ids=companies, user_ids=users)
    print(f"{message}  -- message updated in Notifications table.")


if __name__ == "__main__":
    message = get_message()
    type_ = args_read("type") or "default"
    duration = args_read("duration") or 0  # if no value is passed use 0
    duration = int(duration) or None       # convert 0 to None to get a toast for forever until closed
    horizontal = args_read("horizontal") or "left"
    vertical = args_read("vertical") or "top"
    companies = get_company()
    users = get_user()
    update_notification(
        message=message, type=type_, duration=duration, horizontal=horizontal,
        vertical=vertical, companies=companies, users=users
    )



import os
import logging
from dataclasses import dataclass


@dataclass
class GlobalConstants:
    loggerName: str = "PoLZy"
    dateFormat: str = "%d-%m-%Y"
    dateFormatLong: str = "%Y%m%d_%H%M%S%f"
    filesPath: str = "BatchFiles/"
    archivedFiles: str = os.path.join(filesPath, "processed/")


logger = logging.getLogger(GlobalConstants.loggerName)



from dataclasses import dataclass, field
from polzyFunctions.GlobalConstants import logger
from datetime import datetime
import json


class InputFieldTypes:
    """
    These are the input field types, which are used throughout the application. Each inputfieldVisibilityType triggers
    different behavior in the frontend and has different abilities, additional attributes, etc.

    """
    # Just a regular string. May have valueRangeInput, then it is a dropdown. Otherwise only text field.
    # If used with FieldType.output it may have HTML in the value and it will be rendered accordingly.
    TYPESTRING = "Text"

    # A checkbox, which is displayed on the top of the current card or sub-card
    TYPEBOOLEAN = "Flag"
    # A date with date-selector
    TYPEDATETIME = "Datum"
    # A number.
    # May have valueRangeInput=["range", <low>, <high>]
    # If a range is given and inputTriggersComplexupdates = True it will fire a backend-event
    #     after the value was entered.
    # May also be combined with "decimals" and an int of decimal places. Default = 0 decimals = Integers only.
    TYPENUMERIC = "Zahl"
    # A search field for Address endpoint. Seems to be old and replaced by SearchEndPoint
    TYPEADDRESS = "Adresse"
    # A field that will trigger an endpoint and return with a dropdown list. The search field will consume a full line.
    TYPESEARCH = "SearchEndPoint"
    # Long text field
    TYPETEXTBOX = "TextBox"
    # A grid display with Headers and Lineitems
    TYPETABLE = "Table"
    # An Image with interactive (mouseover) capabilities. Don't mix that up with icons or logos, etc.
    TYPEIMAGE = "Image"

    # Like the TYPEBOOLEAN but will consume one full line of the display.
    # You can provide "related_fields" with a list of (slightly differntly) formatted DataFields (as DICT).
    # The problem with those fields is, that they get sent in the activity as regular fields and need to be treated
    # specially in the updateFieldValues-implementation of the corresponding class in order for the sent values
    # to find their way back into the related_fields
    # In the attribute "kurztext" (brief) you may provide valid HTML (e.g. used for longer texts
    TYPEFLAGFULLLINE = "FlagWithOptions"

    # Like the TypeFlagFULLINE but as toggle field with 3 possible values: yes/no/none.
    TYPERADIOWITHOPTIONS = "RadioFlagWithOptions"

    # a Chart-display with various options in value (as DICT)
    # <axis.i.name>: <value> <axis.i.unit>
    # data payload should be a list of lists [x, y]:
    # "values": [
    #        [0, 2115584],
    #        [29, 2263761] ]
    # {
    #         "value": {
    #           "axis": {
    #             "x": {
    #               "name": "Days",
    #               "unit": "d"
    #             },
    #             "y": {
    #               "name": "Amount,
    #               "unit": "$"
    #             }
    #           },
    #           "values": [
    #                [0, 2115584],
    #                [29, 2263761],
    #                ...
    #           ]
    #         }
    #       }
    # Important: Use fieldVisibilityType = FieldVisibilityTypes.output - everything else will not work!
    TYPECHART = "Chart"

    # Specific component used for documents-card of Backend-Antrags
    # fieldVisibilityType may only be 2!
    # value is a list dict objects representing document file {
    # "id": <string: file.id>,
    # "name": <string: filename>,
    # "created": <string: date>,
    # "signed": <string> (for instance: "Yes" or "No"),
    # }
    # See also example in Showcase:
    # lines 175+ of https://gogs.earthsquad.global/athos/PoLZy_Showcase/src/master/antrag.py
    TYPEDOCUMENTS = "Documents"

    # fieldVisibilityType may only be 2!
    # value is a list dict objects representing attached file
    # example in lines 128+ of gogs.earthsquad.global/athos/PoLZy_Showcase/src/master/antrag.py
    # {"id": < string: file.id >,
    # "name": < string: filename >,
    # "created": < string: date >,
    # "type": < string: file.type >,
    # "actions": [ "edit", "delete"] }
    # The supported actions are edit (edit file type) and delete (delete attached document).
    TYPEATTACHMENTS = "Attachments"

    def __init__(self):
        self.types = [
            FieldDataType(typeName=self.TYPESTRING),
            FieldDataType(typeName=self.TYPEBOOLEAN),
            FieldDataType(typeName=self.TYPEDATETIME),
            FieldDataType(typeName=self.TYPENUMERIC),
            FieldDataType(typeName=self.TYPEADDRESS),
            FieldDataType(typeName=self.TYPESEARCH),
            FieldDataType(typeName=self.TYPETEXTBOX),
            FieldDataType(typeName=self.TYPETABLE),
            FieldDataType(typeName=self.TYPEIMAGE),
            FieldDataType(typeName=self.TYPEFLAGFULLLINE),
            FieldDataType(typeName=self.TYPECHART),
            FieldDataType(typeName=self.TYPEDOCUMENTS),
            FieldDataType(typeName=self.TYPEATTACHMENTS),
        ]


@dataclass
class FieldDataType:
    typeName: str

    def toJson(self):
        return self.typeName


@dataclass
class FieldVisibilityTypes:
    """
    Self-explanatory

    """
    # Visible and ready for Input in Frontend
    visible = 1
    # Visible in Frontend. Output only.
    output = 2
    # this field-type will generally not be transferred to the frontend. It will usually not be printed.
    hidden = 3
    # This field-type will not be visible in Frontend but be transferred anyway
    hiddenFrontend = 4


@dataclass
class FieldDefinition:
    name: str = field(default="")
    # This value is a single implementation of usually 2 fields (Value, DefaultValue). When the activity get's executed
    # it's not important whether the user chose the default value or manually selected another value, so 1 field is
    # enough. When the backend wants to set a default value, it writes it in this field for the frontend.
    # Before the frontend can execute the activity, it will write the chosen value from the UI into this field.
    value: any = field(default=None)
    valueTech: any = field(default=None)
    valueOutput: any = field(default=None)
    # The value range is always a list. Either a list with all possible entries or a list with ["range", low, high]
    # for numeric fields. For sliders it is ["slider", low, high].
    # for string field sliders it is ["slider", <value>, <value2>, <value3>".
    # The sliders will be positioned according to order in the list (1st entry left,
    #      2nd entry more right than 1st, last entry right)
    valueRangeInput: list = field(default_factory=list)
    # Please check docu in fieldDataType
    fieldDataType: FieldDataType = field(default=FieldDataType)  # Boolean, String, Numeric, etc.
    # for numeric fields only
    decimalPlaces: int = field(default=0)
    # When this flag is set, the frontend will trigger an immediate backend call if the field value changes.
    inputTriggersComplexUpdates: bool = field(default=False)
    shortDescription: str = field(default="")  # Short description, which will be shown in the UI
    tooltip: str = field(default="")  # Tooltip on mouse-over in the UI
    # If this value is set to True, a dropdown list will not allow values other than the ones from ValueRange
    onlyValuesFromValueRange: bool = field(default=False)
    # If True, Frontend will not allow continuation without this field being entered/selected
    isMandatory: bool = field(default=False)
    errorMessage: str = field(default="")  # If a value is set, the frontend will show it as helper to the user.
    fieldVisibilityType: int = field(default=FieldVisibilityTypes.visible)  # See Class FieldVisibilityTypes
    icon: str = field(default="")
    # if the field shall be part of a card = Group, then enter the group-name here. Groups must be defined first
    # using isGroupField=True. The field.name of the group shall go into attribute "group"
    group: str = field(default="")
    # If this is True, then the field is a group-field. Must be boolean.
    # Group-Fields are displayed on the top of the card. They activate/
    # deactivate groups of fields = a card within the card.
    isGroupField: bool = field(default=False)
    # If the value list is too long we can call this endpoint when a value is entered. This is used for instance
    # for address fields to check and provide a value list once the user starts to enter values.
    endpoint: str = field(default="")
    # To group cards further we can create "subtitles" (That are sections within the card). Name here, which subsections
    # you want. Obviously we want to have this attribute only in fields that have also "isGroupField".
    # e.g. ["first section"]
    subtitles: list = field(default_factory=list)
    # If a field shall go into a subsection of the card (see previous attribute) state here the exact name
    # of the section, e.g. "first section" (if you created "first section" within the subtitles of the group.
    subsection: str = field(default="")
    # Option to set colors for the card. Only works with isGroupField=True
    background: str = field(default="")
    # For several FieldDataTypes it is possible to have Child-fields, that will appear only if the host-field has
    # value "true". Here you can pass those additional fields as list.
    relatedFields: list = field(default_factory=list)

    def __post_init__(self):
        pass

    @classmethod
    def stringify(cls, value):
        #
        # stringifies value
        #

        # datetime
        if isinstance(value, datetime):
            return value.strftime("%d-%m-%Y")

        # list or dictionary
        if isinstance(value, (list, dict)):
            return value

        if value is None:
            return value

        # other cases
        return str(value)

    def toJSON(self):
        """
        Generate JSON-String to be sent to Frontend
        :return:
        """

        if hasattr(self, "valueRangeInput") and self.valueRangeInput:
            inputRangeList = [self.stringify(v) for v in self.valueRangeInput]
        else:
            inputRangeList = []

        lValue = self.stringify(self.value)

        # boolean fields with "output" don't work on the frontend. change them to "string" and use valueOutput
        lFieldDataType = self.fieldDataType
        if lFieldDataType == FieldDataType(InputFieldTypes.TYPEBOOLEAN) and self.fieldVisibilityType == FieldVisibilityTypes.output:
            lFieldDataType = FieldDataType(InputFieldTypes.TYPESTRING)
            lValue = self.stringify(self.valueOutput)

        try:
            lReturn = {
                'fieldVisibilityType': self.fieldVisibilityType,
                'name': self.name,
                'brief': self.shortDescription,
                'tooltip': self.tooltip,
                'icon': self.icon,
                'fieldDataType': lFieldDataType.toJson(),
                'inputRange': inputRangeList,
                'onlyFromRange': self.onlyValuesFromValueRange,
                'value': lValue,
                'inputTriggers': self.inputTriggersComplexUpdates,
                'isMandatory': self.isMandatory,
                'errorMessage': self.errorMessage,
                'endpoint': self.endpoint,
                'subtitles': self.subtitles,
                'subsection': self.subsection,
                'backgroundColor': self.background,
                'group': self.group,
                'shortDescription': self.shortDescription,
                'relatedFields': json.loads(json.dumps(
                    self.relatedFields, default=lambda o: o.toJson() if isinstance(o, FieldDataType) else o)
                ),  # converting FieldDataType object to string, just to avoid errors in json.dumps
            }

            # Remove all empty field values to save space for talking to the frontend
            return {k: v for k, v in lReturn.items() if not self.__checkIfEmptyValue(v)}

        except Exception as e:
            logger.critical(f"Fehler {e}")
            raise ValueError(f"Fehler in toJSON: {e}")

    @staticmethod
    def __checkIfEmptyValue(inValue) -> bool:  # Will return true if empty, flase if not empty.
        if not inValue:
            return True
        if isinstance(inValue, str):
            if len(inValue.strip()) > 0:
                return False
            return True
        if isinstance(inValue, (int, float)):
            return False
        if isinstance(inValue, (list, dict)):
            return len(inValue) == 0
        if isinstance(inValue, FieldVisibilityTypes):
            return False

        logger.info(f"not sure how to handle this field value {inValue}, it's type {type(inValue)}. Will parse to FE")
        return True


class InputFields:
    def __init__(self):
        self.fields: [FieldDefinition] = []

    def addField(self, inputField: FieldDefinition):
        self.fields.append(inputField)

    def getAllInputFields(self) -> [FieldDefinition]:
        return self.fields

    def getAllFieldGroups(self) -> [FieldDefinition]:
        return list(filter(lambda field: field.isGroupField, self.fields))

    def __len__(self):
        return len(self.fields)

    def toJSON(self):
        if not hasattr(self, "fields") or not self.fields:
            return []
        outputList = [f.toJSON() for f in self.fields]
        return outputList

    def getField(self, **kwargs) -> FieldDefinition:
        # returns InputField object matching with given parameters.
        for item in self.fields:
            success = True
            for key, value in kwargs.items():
                try:
                    original_value = getattr(item, key)
                except Exception:
                    original_value = ""
                if original_value != value:
                    success = False
                    break
            if success:
                return item
        return None

    def getFieldsByFieldName(self, FieldName="name", value=None):
        return [field_ for field_ in self.fields if getattr(field_, FieldName) == value]

    def setField(self, fieldDefinition: FieldDefinition):
        for item in self.fields:
            if item.name == fieldDefinition.name:
                item.fieldVisibilityType = fieldDefinition.fieldVisibilityType
                item.shortDescription = fieldDefinition.shortDescription
                item.isMandatory = fieldDefinition.isMandatory
                item.feldArt = fieldDefinition.fieldDataType
                item.value = fieldDefinition.value
                item.valueTech = fieldDefinition.valueTech
                item.onlyValuesFromValueRange = fieldDefinition.onlyValuesFromValueRange
                item.inputTriggersComplexUpdates = fieldDefinition.inputTriggersComplexUpdates
                item.valueRange = fieldDefinition.valueRangeInput
                item.tooltip = fieldDefinition.tooltip
        # Looks like a new item!
        self.addField(inputField=fieldDefinition)



import json
import codecs
import os
import sys
import getopt


class ManageTranslator:
    def __init__(self):
        """
        This class is used to manage and maintain translation data for fasifu.
        """
        self.dataFilePath = os.path.join(os.path.dirname(os.path.abspath(__file__)), "language_data.json")
        self.data = self.read_data()
        self.languages = self.data[""].copy()
        del self.data[""]

    def read_data(self):
        # read data with exact encodings using codec.open
        with codecs.open(self.dataFilePath, "r", encoding="utf8")as file:
            return json.load(file)

    def print_table(self):
        """
        This function print all translated data in table form which looks neat
        :return:
        """
        len_dic = {"en": 0}  # stores len of biggest string in each language so table can be made of correct size
        for key in self.data:  # iterates through every string of each language and store the biggest number in dict
            if len(key) > len_dic["en"]:
                len_dic["en"] = len(key)
            for lang in self.data[key]:
                if lang not in len_dic:
                    len_dic[lang] = 0
                if len(self.data[key][lang]) > len_dic[lang]:
                    len_dic[lang] = len(self.data[key][lang])
        format_string = " | ".join([("{:<" + str(len_dic[lang])) + "}" for lang in len_dic])  # manages space
        print(format_string.format(*len_dic.keys()))  # prints header
        print(format_string.format(*["="*len_dic[lang] for lang in len_dic]))  # seperator
        for key, value in self.data.items():
            data = [key] + list(value.values())
            print(format_string.format(*data))  # prints translation values
        print(format_string.format(*["=" * len_dic[lang] for lang in len_dic]))

    def add_item(self, english, save=True):
        """
        Adds english word/sentence in translation data. After these we can store the translation data for these
        word/sentence by using add_word.
        :param english: English word/sentence to be removed.
        :return:
        """
        self.data[english] = self.languages
        if save:
            self.save_data()

    def remove_item(self, english, save=True):
        """
        Removes english word and translations of it from data.
        :param english: English word/sentence to be removed.
        :return:
        """
        try:
            del self.data[english]
            if save:
                self.save_data()
        except Exception as ex:
            print(ex)

    def add_language(self, language_code, save=True):
        """
        Adds a language in data with default value = "" for all english words
        :param language_code: e.g.- English = en, German = de
        :return:
        """
        language_code = language_code.lower()
        for key in self.data:
            if language_code not in self.data[key]:
                self.data[key][language_code] = ""
        self.languages[language_code] = ""
        if save:
            self.save_data()

    def remove_language(self, language_code, save=True):
        """
        removes translation of a language for all words
        :param language_code: e.g.- English = en, German = de
        :return:
        """
        language_code = language_code.lower()
        for key in self.data:
            del self.data[key][language_code]
        if language_code in self.languages:
            del self.languages[language_code]
        if save:
            self.save_data()

    def add_translation(self, english, language_code, translated_word, save=True):
        """
        Adds new translation for the supplied english word.
        :param english: English word whose translation is to be supplied.
        :param language_code: Language code of the language whose translation is to be supplied. e.g.- English = en
        :param translated_word: Translation of the english word in the suppied language
        :return:
        """
        language_code = language_code.lower()
        if language_code not in self.languages:
            self.add_language(language_code)
        if not self.data.get(english):
            self.add_item(english)
        self.data[english][language_code] = translated_word
        if save:
            self.save_data()

    def get_empty_by_language_code(self, language_code):
        """
        Returns list of english words which doesn't have translation for the supplied language_code.
        :param language_code: e.g.- English = en, German = de
        :return: List
        """
        language_code = language_code.lower()
        data = []
        for key in self.data:
            if not self.data[key][language_code]:
                data.append(key)
        return data

    def fill_empty_by_language_code(self, language_code, save=True):
        """
        Fill translated data one by one for the english words doesn't have translation for the supplied language_code.
        :param language_code: e.g.- English = en, German = de
        :return:
        """
        language_code = language_code.lower()
        data = self.get_empty_by_language_code(language_code)
        for english in data:
            print("en:", english)
            translation = input(f"{language_code}: ")
            self.add_translation(english, language_code, translation, save)
            print("added.\n\n")

    def get_all_empty(self):
        """
        Returns List of lists. These lists contain english word/sentence and language code which doesn't have translation
        for the supplied english word/sentence.
        :return:
        """
        data = []
        for key in self.data:
            for language_code in self.languages:
                if not self.data[key][language_code]:
                    data.append([key, language_code])
        return data

    def fill_all_empty(self, save=True):
        """
        Prints english word which doesn't has translation for the language code below it. You can fill all the
        translation one by one.
        :return:
        """
        data = self.get_all_empty()
        for english, language_code in data:
            print("en:", english)
            translation = input(f"{language_code}: ")
            self.add_translation(english, language_code, translation, save)
            print("added.\n\n")

    def save_data(self):
        self.data[""] = self.languages
        with codecs.open(self.dataFilePath, "wb", encoding="utf8")as file:
            json.dump(self.data, file, indent=4, ensure_ascii=False)  # writes data and character perfectly


def print_args():
    print("""
    Call: python ManageTranslations.py --parameters 
           --print_table: Prints table of all translations with language code as column name.
           --add_item <english word/sentence to be added in translation>: Adds new word in translations.
           --remove_item <english word/sentence to be removed from translation>: Removes word from translations.
           --add_language <language code e.g.- For English = en>: Adds column for new language code in preexisting and new data
           --remove_language <language code e.g.- For English = en>: Removes column form preexisting and new data
           --add_translation -e <english version of the translation> -l <language_code of translation> -t <translated>: Add/change translation
           --fill_empty_by_language <language_code>: Get none translated words for supplied language and fill translation one by one
           --fill_empty: Get none translated words from all data and fill translation one by one
       """)


def args_read(l_search_parameter):
    l_args = sys.argv[1:]

    try:
        opts, args = getopt.getopt(l_args, "e:l:t:", ["print_table",
                                                "add_item=",
                                                "remove_item=",
                                                "add_language=",
                                                "remove_language=",
                                                "add_translation",
                                                "fill_empty_by_language=",
                                                "fill_empty",
                                                ])
    except getopt.GetoptError as err_det:
        print("Error in reading parameters:" + str(err_det))
        sys.exit("Wrong parameters - exiting")
    if opts:
        for opt, arg in opts:
            if l_search_parameter == opt:  # in ("-u", "--usage"):
                return arg
            if "--" + l_search_parameter == opt or "-" + l_search_parameter == opt:
                if arg:
                    return arg
                elif opt in ["--print_table", "--add_translation", "--fill_empty"]:
                    return opt
    return None

def manage_args():
    m = ManageTranslator()
    if args_read("print_table"):
        m.print_table()
    elif args_read("add_item"):
        m.add_item(args_read("add_item"))
    elif args_read("remove_item"):
        m.remove_item(args_read("remove_item"))
    elif args_read("add_language"):
        m.add_language(args_read("add_language"))
    elif args_read("remove_language"):
        m.remove_language(args_read("remove_language"))
    elif args_read("add_translation"):
        m.add_translation(args_read("e"), args_read("l"), args_read("t"))
    elif args_read("fill_empty_by_language"):
        m.fill_empty_by_language_code(args_read("fill_empty_by_language"))
    elif args_read("fill_empty"):
        m.fill_all_empty()
    else:
        print_args()


if __name__ == "__main__":
    manage_args()


import os
import json
import codecs
import threading
from polzyFunctions.GlobalConstants import logger

lock = threading.Lock()  # to make class thread safe


class Singleton(type):
    _instances = {}

    def __call__(cls, *args, **kwargs):
        if cls not in cls._instances:
            with lock:
                if cls not in cls._instances:
                    cls._instances[cls] = super(Singleton, cls).__call__(*args, **kwargs)
        return cls._instances[cls]


class Data(metaclass=Singleton):
    """
    This is class is used to store language data. It was made to enhance performance.
    This is a singleton and thread safe class, because of this it loads json file
    only when the class is first initiated, later all the time it returns the same data.
    This also means that any changes made on language_data.json on runtime will not affect the program until restart.
    """
    def __init__(self):
        logger.debug("Language translation backend module initialized - buffered")
        self.data = self.get_data(os.path.join(os.path.dirname(os.path.abspath(__file__)), "language_data.json"))

    @staticmethod
    def get_data(fileNameAndPath):
        with codecs.open(fileNameAndPath, 'r') as file:
            data = json.load(file)
        return data



from polzyFunctions.GlobalConstants import logger
from polzybackend.models import AntragActivityRecords


def recordAntragDecorator(method):
    """
    Decorator to write the current status after the activity was execute to Database.
    :param method:
    :return:
    """
    def wrapper(self, *args, **kwargs):
        result = method(self, *args, **kwargs)
        if hasattr(self, "antrag"):
            self.antrag.latestDBTimestamp = activityWriter(self)
        else:
            self.latestDBTimestamp = activityWriter(self)
        return result
    return wrapper


def activityWriter(inInstance):
    """
    The usage of "Antrag" here is wrong, usually it's an AntragActivitiy that will be provided in the call.
    It is anyway important to have this reference because otherwise the "where used" doesn't work for "setSearchString"

    """
    from polzyFunctions.Dataclasses.Antrag import Antrag

    lInstance = None
    try:
        lInstance = inInstance.antrag
    except AttributeError:
        pass

    if not lInstance:
        # this will work if the instance itself is an Antrag (happens so far only when the Tag is changed by
        # frontend
        lInstance:Antrag = inInstance

    logger.debug(f"classname = {lInstance.__class__.__name__}")
    # Update search string in this antrag instance. It is the main search criteria.
    try:
        lInstance.setSearchString()
    except AttributeError:
        logger.critical(f"We have instance of Type {type(inInstance)}. We can't currently work with that.")
        return

    record = AntragActivityRecords.new(lInstance)
    logger.debug(f"Activity recorded with id: {record.id}")
    return record.timestamp



from flask import jsonify, request, current_app, send_file, abort
from polzybackend.antrag import bp
from polzybackend.utils.import_utils import antrag_products, antrag_class
from polzybackend import auth
from polzybackend.models import AntragActivityRecords
from polzybackend.utils.import_utils import import_class
import json

@bp.route('/antrag/products')
@auth.login_required
def get_antrag_products():
    #
    # returns all products that available for current user
    #
    try:
        # get all stages
        product_list = antrag_products().getAllAvailableProducts(auth.current_user())
        return jsonify(product_list), 200

    except Exception as e:
        current_app.logger.exception(f"Error during get_antrag_products: {e}")
        
    return jsonify({'error': f'Failed to get antrag products'}), 400


@bp.route('/antrag/new/<string:product_type>')
@auth.login_required
def new_antrag(product_type):
    #
    # creates an Antrag by product_type
    # and returns it to frontend
    #
    try:
        antrag = antrag_class()(product_type, auth.current_user())

        # store antrag and return it to store and return json object
        antrag.initialize()
        current_app.config['ANTRAGS'][antrag.id] = antrag
        result = antrag.get()
        return jsonify(result), 200

    except Exception as e:
        current_app.logger.exception(f'Initialization of antrag instance {product_type} failed: {e}')

    return jsonify({'error': f'Initialization of antrag instance failed'}), 400


@bp.route('/antrag/clone/<string:id>')
@auth.login_required
def clone_antrag(id):
    #
    # clone antrag instance by id
    #

    try:
        # get antrag from app store
        antrag_src = current_app.config['ANTRAGS'].get(id)
        if antrag_src is None:
            raise Exception(f'Antrag {id} not found')

        # make copy and return antrag json object
        antrag = antrag_src.clone()
        current_app.config['ANTRAGS'][antrag.id] = antrag
        result = antrag.get()
        return jsonify(result), 200

    except Exception as e:
        current_app.logger.warning(f'Cloning of antrag {id} failed: {e}')
    
    return jsonify({'error': f'Cloning of antrag instance failed'}), 400


@bp.route('/antrag/delete/<string:id>', methods=['DELETE'])
@auth.login_required
def delete_antrag(id):
    #
    # delete antrag instance from store by <id>
    #

    # check if instance exists
    if id not in current_app.config['ANTRAGS']:
        return jsonify({'error': f'Antrag instance {id} not found'}), 404

    current_app.config['ANTRAGS'] = {key: value for key, value in current_app.config['ANTRAGS'].items() if key != id}
    return {'OK': f'Antrag instance {id} successfully deleted'}, 200



@bp.route('/antrag/tag/<string:antrag_id>', methods=['POST', 'DELETE'])
@auth.login_required
def antrag_tag(antrag_id):
    #
    # set or delete custom tag of antrag
    #

    # get antrag from app store
    antrag = current_app.config['ANTRAGS'].get(antrag_id)
    if antrag is None:
        current_app.logger.warning(f'Antrag {antrag_id} is absent in PoLZy storage. Most probably instance restarted.')
        return jsonify({'error': f'Antrag instance not found'}), 400

    # update tag
    if request.method == 'POST':
        # get tag from payload
        tag = request.get_json().get('tag')
        if tag:
            antrag.instance.setCustomTag(tag)
            return {'OK': 'Custom Tag successfully set'}, 200

    # delete tag
    antrag.instance.setCustomTag(None)
    return {'OK': 'Custom Tag successfully deleted'}, 200


@bp.route('/antrag/update', methods=['POST'])
@auth.login_required
def update_antrag():
    #
    # updates antrag fields
    #

    # get post data
    data = request.get_json()

    # get antrag and update its values
    try:
        # get antrag from app store
        antrag = current_app.config['ANTRAGS'].get(data['id'])
        if antrag is None:
            raise Exception(f'Antrag {data["id"]} is absent in PoLZy storage. Most probably instance restarted.')

        # update antrag values and return antrag json object
        antrag.updateFields(data)
        result = antrag.get()
        return jsonify(result), 200

    except Exception as e:
        current_app.logger.exception(f'Antrag {data["id"]}, fields update failed: {e}')
    
    return jsonify({'error': f'Update of the antrag fields failed'}), 400


@bp.route('/antrag/execute', methods=['POST'])
@auth.login_required
def execute_antrag():
    #
    # executes antrag activities
    #

    # get post data
    data = request.get_json()

    # get antrag and execute activity
    try:
        # get antrag from app store
        antrag = current_app.config['ANTRAGS'].get(data['id'])
        if antrag is None:
            raise Exception(f'Antrag {data["id"]} is absent in PoLZy storage. Most probably instance restarted.')

        # execute antrag activity and return the response object
        result = antrag.executeActivity(data)
        return jsonify(result), 200

    except Exception as e:
        current_app.logger.exception(f'Failed execute activity {data.get("activity")} of antrag {data["id"]}: {e}')
    
    return jsonify({'error': f'Execution of antrag activiy {data.get("activity")} failed'}), 400



@bp.route('/antrag/records/search', methods=['POST'])
@auth.login_required
def getSearchStringFromRecords():
    data = request.get_json()

    # supplying current user to get records of current user & company
    found_antrags = AntragActivityRecords.getSearchString(auth.current_user(), data.get("value"))
    results = [{
        'id': instance.antrag_id,
        'label': instance.get_label(),  # 'get_label' method should be adjusted to proper render results
    } for instance in found_antrags] if found_antrags else []

    return jsonify(results), 200

# we have to pass only antrag id. it will be easier to use GET method
@bp.route('/antrag/records/<string:antrag_id>')#load', methods=['POST'])
@auth.login_required
def loadLatestRecords(antrag_id):

    # preventing duplication of the antrag instance
    if antrag_id in current_app.config['ANTRAGS']:
        return {'error': 'Antrag is already active'}, 409

    # get antrag record by id
    antrag_record = AntragActivityRecords.getLatest(antrag_id)
    if antrag_record is None:
        return {'error': f'No record found of antrag {antrag_id}'}, 404

    # create antrag instance from the record
    antrag = antrag_class()(
        antrag_record.class_name,
        auth.current_user(),
        id=antrag_id,
    )

    # load antrag instance and store it within the app
    antrag.load(antrag_record.sapClient)
    current_app.config['ANTRAGS'][antrag.id] = antrag

    # creating dictionary with name as key and value as value of inputField. These are used to load fields.
    fields_json_data = {js.get("name"): js.get("value") for js in antrag_record.json_data}
    sapAttributes_json_data = {js.get("name"): js.get("sapFieldAttributes") for js in antrag_record.json_data}

    # update field values from the record and return the result
    antrag.instance.id = antrag_record.antrag_id  # using same antrag_id as from record to avoid new record because of
    antrag.instance.updateFieldValuesFromDatabase(fields_json_data)                                   ## new antrag id
    antrag.instance.updateSapAttributesFromDatabase(sapAttributes_json_data)
    antrag.instance.status = antrag_record.status
    antrag.instance.antragsnummer = antrag_record.antragsnummer
    antrag.instance.latestDBTimestamp = antrag_record.timestamp
    try:
        antrag.instance.productName = antrag.instance._generateProduktNameWithAntragsnummer()
    except:
        pass
    antrag.instance.loadActivitiesFromDict(antrag_record.json_data_activities)
    # update tag
    antrag.instance.setCustomTag(antrag_record.tag)
    antrag.instance.fillCurrentlyPossibleActivities()
    result = antrag.get()
    return jsonify(result), 200




from polzybackend.models import GamificationActivity
from polzyFunctions.Activities.LoginActivity import LoginActivity
import json


def recordActivityDecorator(method):
    """
    This is a decorator that will write into GamificationActivity table.
    :param method: The method to be executed
    :return:
    """
    def wrapper(self, *args, **kwargs):
        writer = WriteActivity(self)
        writer.activityWriter()
        return method(self, *args, **kwargs)

    return wrapper


class WriteActivity:
    """
    Writes an activity to gamification table.
    """
    def __init__(self, instanceOfAntragPolicyOrActivity):
        from polzyFunctions.Activities.AntragActivity import AntragActivity
        from polzyFunctions.Activities.Activity import Activity
        from polzyFunctions.Dataclasses.Antrag import Antrag
        # FIXME: Fix after policy refactor
        #from fasifu.Activities.PolicyActivity import PolicyActivity
        #from fasifu.Dataclasses.Polizze import Polizze

        self.activity: Activity = None
        self.antrag: Antrag = None
        self.polizze = None
        self.login: LoginActivity = None

        self.callingClassRef = instanceOfAntragPolicyOrActivity

        if isinstance(instanceOfAntragPolicyOrActivity, AntragActivity):
            self.activity = instanceOfAntragPolicyOrActivity
            self.event = 4  # "AntragActivity"
        elif "PolicyActivity" in instanceOfAntragPolicyOrActivity.__class__.__name__:
            self.activity = instanceOfAntragPolicyOrActivity
            self.event = 3  # "PolicyActivity"
        elif isinstance(instanceOfAntragPolicyOrActivity, Antrag):
            self.antrag = instanceOfAntragPolicyOrActivity
            self.event = 1  # "Antrag"
        elif "Polizze" in instanceOfAntragPolicyOrActivity.__class__.__name__:
            self.polizze = instanceOfAntragPolicyOrActivity
            self.event = 2  # "Polizze"
        elif isinstance(instanceOfAntragPolicyOrActivity, LoginActivity):
            self.login = instanceOfAntragPolicyOrActivity
            self.event = 6

    def activityWriter(self):
        user = None
        eventDetails = None

        if self.polizze:
            try:
                user = self.polizze.user
            except AttributeError:
                # This is most probably an error record, e.g. because the policy doesn' exist.
                return
            eventDetails = self.__getEventDetailsPolizze()
        elif self.antrag:
            user = self.antrag.user
            eventDetails = self.__getEventDetailsAntrag()
        elif self.activity:
            try:
                user = self.activity.antrag.user
                self.antrag = self.activity.antrag
                eventDetails = self.__getEventDetailsAntrag()
            except:
                user = self.activity.polizze.user
                self.polizze = self.activity.polizze
                eventDetails = self.__getEventDetailsPolizze()
            eventDetails.update({"Activity": self.callingClassRef.__class__.__name__})
        elif self.login:
            user = self.login.user
            eventDetails = self.login.event_details

        GamificationActivity.new(user=user, event=self.event, event_details=json.dumps(eventDetails))

    def __getEventDetailsPolizze(self):
        eventDetails = {"polizzenNummer": self.polizze.polizzenNummer,
                        "stage": self.polizze.user.stage,
                        "sapClient": self.polizze.sapClient,
                        "lineOfBusiness": self.polizze.lineOfBusiness,
                        "productName": self.polizze.productName}
        return eventDetails

    def __getEventDetailsAntrag(self):
        eventDetails = {"lineOfBusiness": self.antrag.lineOfBusiness,
                        "stage": self.antrag.user.stage,
                        "sapClient": self.antrag.sapClient,
                        "productName": self.antrag.productName}
        return eventDetails



from polzyFunctions.Activities.ActivitiesDataClasses import FieldDefinition, InputFieldTypes, FieldDataType, \
    FieldVisibilityTypes
from polzyFunctions.GlobalConstants import logger
from datetime import datetime   # Needed because dynamically called
from polzyFunctions.translator import Translator
from polzyFunctions.ConfigurationEngine.ConfigurationProvider import lConfigurationProvider
from polzyFunctions.Dataclasses.CommonFieldnames import CommonFieldnames
import codecs
import json
import os

lTranslator = Translator(default=lConfigurationProvider.language)  # <-- This will lead to English.


def get_field_catalog_dictionary(JsonFilesDirectory):
    """
    This function is used to get dictionary based on file/function name. Fields can be stored in JSON-Files
    in the form:
    [
    {
        "fieldVisibilityType": "$(FieldVisibilityTypes.visible)",
        "name": "$(CommonFieldnames.policyBeginDate.value)",
        "tooltip": "Policy beginn date ",
        "shortDescription": "Policy begin date",
        "fieldDataType": "$(FieldDataType(typeName=InputFieldTypes.TYPEDATE))"
    },
    ...
    ]

    These JSON-files can have the same name as each activity.

    :return:
    """

    field_data = dict()
    if JsonFilesDirectory is None:
        return

    filenames = os.listdir(JsonFilesDirectory)
    for filename in filenames:
        if filename[-5:].lower() != ".json":
            continue
        json_file = os.path.join(JsonFilesDirectory, filename)
        with codecs.open(json_file, "r", encoding="utf-8") as file:  # using codecs with utf-8 encoding to avoid errors
            dic = json.load(file)
        field_data[filename[:-5]] = dic  # storing key as filename without extension in order to match function name
    return field_data


class ManageFieldCatalog:
    """
    Here are all field definitions for all policy activities and types of quotations

    The fields are loaded from JSON-files, which have the same name as the Antrag or Activity.
    The are in ../Dataclasses/JsonFiles. See method updateFieldDefinition for details.

    If you want to debug the processing of a JSON-file, put a key "breakpoint" with any value into the JSON and
    we'll issue a breakpoint-statement.
    """

    # This class attribute is necessary and you must call it in sub class and supply FieldCatalog Json Files directory
    # as argument for "JsonFilesDirectory". This method will return a dictionary containing FieldCatalog data for all
    # activities

    # getting dictionary of FieldCatalog data for all functions
    field_data = get_field_catalog_dictionary(JsonFilesDirectory=None)  # None supplied only for making dummy

    @staticmethod
    def addFields(self):
        if not hasattr(ManageFieldCatalog, self.__class__.__name__):
            logger.debug(f"There is no field list for {self.__class__.__name__}. Most probably OK")
            return

        ManageFieldCatalog.determineLanguage(self)  # set Translator to proper Language

        # Execute filling the field catalog for either the antrag, Policy or Activity.
        try:
            getattr(ManageFieldCatalog, self.__class__.__name__)(self)
            logger.debug(f"FielCatalog created for {self.__class__.__name__}")
        except AttributeError as e:
            logger.exception(f"Building field catalog for {self.__class__.__name__} failed. Error was: {e}")

        # Now, that the field-catalog is built, check if we have default values from
        self.deriveDefaultsFromUserOrCompany()

    @staticmethod
    def determineLanguage(self):
        """
        Try to get the user's language for the translator based on User-Object, which is in Antrag or via
        Activity.antrag. If found, we set it as translation language, otherwise we let it in english.
        :param self:
        :return:
        """
        global lTranslator


        desiredLanguage = None
        try:
            desiredLanguage = self.antrag.user.language
        except AttributeError:
            pass

        if not desiredLanguage:
            try:
                desiredLanguage = self.user.language
            except AttributeError:
                pass

        if desiredLanguage:
            lTranslator = Translator(default=desiredLanguage)

    @staticmethod
    def evaluate_value(value, self):
        """
        this function converts string to python literal and treat it as statement. It is used in FieldCatalog values.
        :param value: a value, that can be evaluated to an expression, e.g. "$(True)" will evaluate to bool True.
        :param self: the current instance (of an activity or antrag)
        :return:
        """

        if isinstance(value, dict):  # if it is a dict we need to convert it to json
            for key_, value_ in value.items():
                value[key_] = ManageFieldCatalog.evaluate_value(value_, self)
            value = json.dumps(value)

        # if it has $ prefix with parenthesis in starting then it is a python statement, evaluate it and return it.
        if isinstance(value, str) and len(value) >= 4 and str(value)[:2] + str(value)[-1] == "$()":
            value = eval(value[2:-1])
        return value

    @classmethod
    def updateFieldDefinition(cls, name, self, keys=None, **kwargs):
        """

        :param name: Name of the JSON-File to load the field definition from (this was already cached during init)
        :param self: The object (Either an Antrag-instance or an Activity-Instance)
        :param keys: Keys to load from the JSON. If not provided all records will be loaded
        :param kwargs: Additional arguments, that will be included in the resulting fields, e.g. "backgroundcolor"
        :return:
        """
        field_data_list = cls.field_data.get(name)  # getting list of fieldCatalog data by function name
        for key, value in kwargs.items():  # some functions needs various variables in arguments. These variables are
            globals()[key] = value  # not same always. So, for flexibility we are taking kwargs input and make
        globals()["self"] = self  # them global variable. This is mainly used while evaluating string statement
        # as python literal.

        for data in field_data_list:
            if "breakpoint" in data.keys():
                breakpoint()
                del data["breakpoint"]
            if keys:  # if keys list is supplied that means we need to update FieldCatalog only for those fields.
                field_name = cls.evaluate_value(data.get("name"), self)  # getting evaluated value for name
                if field_name not in keys:
                    continue
            dic = {}
            for key, value in data.items():
                evaluated_value = cls.evaluate_value(value, self)  # evaluating value
                dic[key] = evaluated_value
            self._addFieldForActivity(FieldDefinition(**dic))  # supplying dictionary of evaluated values as kwargs
        for key in kwargs.keys():  # removing globals variable we set above
            del globals()[key]

    @staticmethod
    def AntragCommon(self):
        """
        Common fields for all Antrag instances
        :param self:
        :return:
        """
        ManageFieldCatalog.updateFieldDefinition(name="AntragCommon", self=self)



if __name__ == "__main__":
    import AddRoot
import io
import sys
from polzybackend import models
from polzybackend import create_app
from flask_sqlalchemy import SQLAlchemy


save_stdout = sys.stdout  # saves original stdout
sys.stdout = io.StringIO()  # using dummy handler for stdout
from polzyFunctions.scripts.config import Config  # This command throws a print statement therefore using dummy stdout handler
sys.stdout = save_stdout  # restoring original stdout handler


def get_db():
    Config.DEBUG = True
    app = create_app(Config)
    return SQLAlchemy(app)

db = get_db()
models.db = db  # replacing module's db with initiatized db so session commits go successful


admin = db.session.query(models.User).filter_by(email="admin@polzy.com").first()
sample_organiztion = db.session.query(models.Company).filter_by(name="Sample Organization").first()
if sample_organiztion:
    sample_organiztion_id = sample_organiztion.id



from polzybackend import db, auth
from polzybackend.utils import generate_id, date_format
from polzybackend.utils.auth_utils import generate_token, get_expired, is_supervisor, load_attributes
from polzyFunctions.GlobalConstants import logger
from datetime import datetime, date
from sqlalchemy import and_, or_
from functools import reduce
from ast import literal_eval
import base64
import json
import os


# authentication
@auth.verify_token
def verify_token(token):
    user = User.query.filter_by(access_key=token).first()
    if user and user.key_expired > datetime.utcnow():
        return user


#
# Authentication Models
#

#
# Association Models
#

# association table for user-role relation
user_company_roles = db.Table(
    'user_company_roles',
    db.Column('user_id', db.String(56), db.ForeignKey('users.id'), primary_key=True),
    db.Column('company_id', db.String(56), db.ForeignKey('companies.id'), primary_key=True),
    db.Column('role_id', db.Integer, db.ForeignKey('roles.id'), primary_key=True)
)


class CompanyToCompany(db.Model):
    #
    # association object between companies
    #
    __tablename__ = 'company_company'
    parent_id = db.Column(db.String(56), db.ForeignKey('companies.id'), primary_key=True)
    child_id = db.Column(db.String(56), db.ForeignKey('companies.id'), primary_key=True)
    attributes = db.Column(db.String(1024), nullable=True)

    # relationships
    parent = db.relationship('Company', backref='child_companies', foreign_keys=[parent_id])
    child = db.relationship('Company', backref='parent_companies', foreign_keys=[child_id])


class UserToCompany(db.Model):
    #
    # association object between user and company
    #
    __tablename__ = 'user_company'
    user_id = db.Column(db.String(56), db.ForeignKey('users.id'), primary_key=True)
    company_id = db.Column(db.String(56), db.ForeignKey('companies.id'), primary_key=True)
    attributes = db.Column(db.String(1024), nullable=True)

    # relationships
    user = db.relationship(
        'User',
        lazy='selectin',
        backref='companies',
        foreign_keys=[user_id],
    )
    company = db.relationship(
        'Company',
        lazy='selectin',
        backref='users',
        foreign_keys=[company_id],
    )
    roles = db.relationship(
        'Role',
        lazy='selectin',
        secondary=user_company_roles,
        primaryjoin=and_(user_id == user_company_roles.c.user_id, company_id == user_company_roles.c.company_id),
        backref=db.backref('user_to_companies'),
    )
    #badges = db.relationship(
    #    'GamificationBadge',
    #    lazy='subquery',
    #    primaryjoin="and_(UserToCompany.user_id==GamificationBadge.user_id, "
    #                     "UserToCompany.company_id==GamificationBadge.company_id)"
    #)


    def to_json(self):
        return {
            'id': self.company.id,
            'name': self.company.name,
            'displayedName': str(self.company),
            'attributes': load_attributes(self.company.attributes),
        }

    def to_admin_json(self):
        child_companies = CompanyToCompany.query.filter_by(parent_id=self.company_id).all()
        return {
            'id': self.company.id,
            'name': self.company.name,
            'displayedName': str(self.company),
            'childCompanies': [{
                'id': ch.child_id,
                'name': ch.child.name,
                'attributes': ch.attributes,
            } for ch in child_companies],
            'users': [{
                'id': u.user.id,
                'email': u.user.email,
                'roles': [r.name for r in u.roles],
                'attributes': u.attributes if u.attributes else '',
            } for u in self.company.users],
        }


class Role(db.Model):
    __tablename__ = 'roles'
    id = db.Column(db.Integer, primary_key=True)
    name = db.Column(db.String(16), unique=True, nullable=False)
    is_supervisor = db.Column(db.Boolean, nullable=False, default=False)


class User(db.Model):
    __tablename__ = 'users'
    id = db.Column(db.String(56), primary_key=True, default=generate_id)
    email = db.Column(db.String(64), unique=True, nullable=False)
    created = db.Column(db.DateTime, nullable=False, default=datetime.utcnow)
    displayed_name = db.Column(db.String(64), nullable=True)
    first_name = db.Column(db.String(64), nullable=True)
    last_name = db.Column(db.String(64), nullable=True)
    # oauth_provider_id = db.Column(db.Integer, db.ForeignKey('oauth_providers.id'), nullable=False)
    # oauth_user_id = db.Column(db.String(128), nullable=False)
    # oauth_token = db.Column(db.String(128), nullable=False)
    access_key = db.Column(db.String(128), nullable=False, default=generate_token)
    key_expired = db.Column(db.DateTime, nullable=False, default=get_expired)
    # current session attributes
    stage = db.Column(db.String(8), nullable=True)
    language = db.Column(db.String(8), nullable=True)
    company_id = db.Column(db.String(56), db.ForeignKey('companies.id'), nullable=True)

    # relationships
    # oauth_provider = db.relationship(
    #    'OAuthProvider',
    #    foreign_keys=[oauth_provider_id],
    # )
    
    company = db.relationship(
        'UserToCompany',
        primaryjoin=and_(
            id == UserToCompany.user_id,
            company_id == UserToCompany.company_id,
        ),
        uselist=False,
        lazy='selectin',
        overlaps='companies, user',
    )
    badges = db.relationship(
        'GamificationBadge',
        primaryjoin="and_(User.id==GamificationBadge.user_id, "
                         "User.company_id==GamificationBadge.company_id)",
        overlaps='gamification_badges, user',
    )

    def __str__(self):
        return self.displayed_name or self.first_name or self.last_name or self.email.split('@')[0]

    def set_stage(self, stage):
        self.stage = stage
        try:
            db.session.commit()
        except Exception as ex:
            logger.critical(f"Exception while committing changes to db: {ex}")
            db.session.rollback()

    def set_language(self, language):
        self.language = language
        try:
            db.session.commit()
        except Exception as ex:
            logger.critical(f"Exception while committing changes to db: {ex}")
            db.session.rollback()

    def set_company(self, company_id=None, company=None):
        # check if company data is provided
        if company_id is None and company is None:
            raise Exception('Neither Company ID nor Company provided')

        # check if company should be fetched
        if company is None:
            company = db.session.query(Company).filter_by(id=company_id).first()

        # check if company exists
        if company is None:
            raise Exception('Company not found')

        # get UserToCompany instance
        user_company = db.session.query(UserToCompany).filter(and_(
            UserToCompany.user_id == self.id,
            UserToCompany.company_id == company.id,
        )).first()

        # check if association exists
        if user_company is None:
            raise Exception(f'User is not assigned to company ID {company_id}')

        # set company
        self.company_id = company.id
        try:
            db.session.commit()
        except Exception as ex:
            logger.critical(f"Exception while committing changes to db: {ex}")
            db.session.rollback()

        # return company json
        return user_company.to_json()

    def is_supervisor(self):
        return reduce(lambda result, company: result or is_supervisor(company), self.companies, False)

    def get_permissions(self):
        if self.company:
            user_roles = [r.name.lower() for r in self.company.roles]

            return {
                'policy': any(role in ['admin', 'clerk'] for role in user_roles),
                'antrag': any(role in ['admin', 'agent'] for role in user_roles),
            }

        return {
            'policy': False,
            'antrag': False,
        }

    def to_json(self):
        return {
            'id': self.id,
            'email': self.email,
            'name': str(self),
            'isSupervisor': self.is_supervisor(),
            'stage': self.stage,
            'accessToken': self.access_key,
            'companies': [company.to_json() for company in self.companies],
        }

    @classmethod
    def get_user_by_id(cls, user_id):
        return db.session.query(cls).filter_by(id=user_id).first().to_json()

    def get_admin_json(self):
        return {
            'possibleRoles': [r.name for r in Role.query.all()],
            'companies': [company.to_admin_json() for company in filter(
                lambda company: is_supervisor(company),
                self.companies
            )],
        }


'''
class OAuthProvider(db.Model):
    __tablename__ = 'oauth_providers'
    id = db.Column(db.Integer, primary_key=True)
    name = db.Column(db.String(64), unique=True, nullable=False)
    client_id = db.Column(db.String(128), nullable=False)
    secret_key = db.Column(db.String(128), nullable=False)

    def __str__(self):
        return self.name
'''


#
# Company Model
#
class Company(db.Model):
    __tablename__ = "companies"
    id = db.Column(db.String(56), primary_key=True, default=generate_id)
    created = db.Column(db.DateTime, nullable=False, default=datetime.utcnow)
    name = db.Column(db.String(128), unique=True, nullable=False)
    displayed_name = db.Column(db.String(64), nullable=True)
    email = db.Column(db.String(32), nullable=True)
    phone = db.Column(db.String(16), nullable=True)
    country = db.Column(db.String(32), nullable=True)
    post_code = db.Column(db.String(8), nullable=True)
    city = db.Column(db.String(32), nullable=True)
    address = db.Column(db.String(64), nullable=True)
    attributes = db.Column(db.String(1024), nullable=True)

    def __str__(self):
        return self.displayed_name or self.name


#
# Activity Models
#
class Activity(db.Model):
    __tablename__ = 'activities'
    id = db.Column(db.String(56), primary_key=True, default=generate_id)
    creator_id = db.Column(db.String(56), db.ForeignKey('users.id'), nullable=False)
    created = db.Column(db.DateTime, nullable=False, default=datetime.utcnow)
    policy_number = db.Column(db.String(64), nullable=False)
    effective_date = db.Column(db.Date, nullable=False, default=date.today)
    type = db.Column(db.String(32), nullable=False)
    status = db.Column(db.String(64), nullable=True)
    finished = db.Column(db.DateTime, nullable=True)
    attributes = db.Column(db.String, nullable=True)

    # relationships
    creator = db.relationship(
        'User',
        backref=db.backref('created_activities', order_by='desc(Activity.created)'),
        foreign_keys=[creator_id],
    )

    def __str__(self):
        return self.id

    @classmethod
    def new(cls, data, policy, current_user):
        #
        # create instance using data
        #

        instance = cls(
            policy_number=policy.number,
            effective_date=datetime.strptime(policy.effective_date, date_format).date(),
            type=data['activity'].get('name'),
            creator_id=current_user.id,
            attributes=json.dumps(data['activity'].get('fields'))
        )

        # store to db
        db.session.add(instance)
        try:
            db.session.commit()
        except Exception as ex:
            logger.critical(f"Exception while committing changes to db: {ex}")
            db.session.rollback()

        return instance

    @classmethod
    def read_policy(cls, policy_number, effective_date, current_user):
        #
        # create instance of reading a policy
        #

        instance = cls(
            policy_number=policy_number,
            effective_date=datetime.strptime(effective_date, date_format).date(),
            type='Read Policy',
            creator=current_user,
            finished=datetime.utcnow(),
            status='OK',
        )

        # store to db
        db.session.add(instance)
        try:
            db.session.commit()
        except Exception as ex:
            logger.critical(f"Exception while committing changes to db: {ex}")
            db.session.rollback()

        return instance

    def finish(self, status):
        #
        # sets is_finished to True
        #
        self.status = status
        self.finished = datetime.utcnow()
        try:
            db.session.commit()
        except Exception as ex:
            logger.critical(f"Exception while committing changes to db: {ex}")
            db.session.rollback()


#
# File Model
#
class File(db.Model):
    __tablename__ = 'files'
    id = db.Column(db.String(56), primary_key=True, default=generate_id)
    filename = db.Column(db.String(128), nullable=False)
    created = db.Column(db.DateTime, nullable=False, default=datetime.utcnow)
    user_id = db.Column(db.String(56), db.ForeignKey('users.id'), nullable=False)
    company_id = db.Column(db.String(56), db.ForeignKey('companies.id'), nullable=False)
    processed = db.Column(db.Boolean, nullable=True, default=False)
    status_ok = db.Column(db.Boolean, nullable=True, default=False)
    type = db.Column(db.String(56), nullable=False, server_default="document")
    details = db.Column(db.String(1024), nullable=True, default="{}")
    parent_id = db.Column(db.String(56), nullable=True)

    # relationships
    user = db.relationship('User', backref='files', foreign_keys=[user_id])
    company = db.relationship('Company', backref='files', foreign_keys=[company_id])

    @classmethod
    def new(cls, user, filename, id, parent_id=None, file_type=None):
        #
        # create new instance of File
        #

        instance = cls(
            id=id,
            filename=filename,
            user_id=user.id,
            company_id=user.company_id,
            parent_id=parent_id,
            type=file_type,
        )

        db.session.add(instance)
        try:
            db.session.commit()
        except Exception as ex:
            logger.critical(f"Exception while committing changes to db: {ex}")
            db.session.rollback()

        return instance

    @staticmethod
    def get_attachment(antrag_id):
        return db.session.query(File).filter(File.parent_id == antrag_id).all()

    def set_processed(self, details=None):
        if not details:
            details = {}
        self.processed = True
        self.status_ok = details.get("status_ok", True)
        if isinstance(details, dict):
            details = json.dumps(details)
        self.details = details
        try:
            db.session.commit()
        except Exception as ex:
            logger.critical(f"Exception while committing changes to db: {ex}")
            db.session.rollback()

    def get_current_filename(self):
        original_filename, extension = os.path.splitext(self.filename)
        return self.id + extension


#
# Antrag Model
#
class AntragActivityRecords(db.Model):
    __tablename__ = "antrag_activity_records"
    id = db.Column(db.String(56), primary_key=True, default=generate_id)
    antrag_id = db.Column(db.String(56), primary_key=True)
    user_id = db.Column(db.String(56), db.ForeignKey('users.id'), nullable=False)
    company_id = db.Column(db.String(56), db.ForeignKey('companies.id'), nullable=False)
    antragsnummer = db.Column(db.String(56), nullable=False)
    timestamp = db.Column(db.DateTime, nullable=False, default=datetime.now)
    status = db.Column(db.String(16), nullable=False)
    searchString = db.Column(db.String, nullable=False)
    json_data = db.Column(db.String, nullable=False)
    json_data_activities = db.Column(db.String, default="{}")
    class_name = db.Column(db.String, nullable=False)
    sapClient = db.Column(db.String(16), nullable=False)
    tag = db.Column(db.String, default=None)

    # relationships
    user = db.relationship('User', foreign_keys=[user_id])
    company = db.relationship('Company', foreign_keys=[company_id])

    @classmethod
    def new(cls, antrag):
        json_data = {}
        for activities in antrag.Activities:
            json_data[activities.__class__.__name__] = activities.toJsonForPersistence()
            if activities.encrypt:  # if activity needs to be encrypt then encrypt it and replace
                encrypted_dic = dict()
                for key, value in json_data[activities.__class__.__name__].items():
                    encrypted_dic[cls.encrypt(key)] = cls.encrypt(value)
                encrypted_dic["encrypt"] = True  # used for identifying that this activity needs to be decrypt
                json_data[activities.__class__.__name__] = encrypted_dic

        instance = cls(
            antrag_id=str(antrag.id),
            user_id=str(antrag.user.id),
            company_id=str(antrag.user.company_id),
            antragsnummer=antrag.antragsnummer,
            status=antrag.status,
            searchString=antrag.searchstring,
            json_data=json.dumps(antrag.Fields.toJSON()),
            json_data_activities=json.dumps(json_data),
            class_name=antrag.__class__.__name__,
            sapClient=antrag.sapClient,
            tag=antrag.tag,
        )
        db.session.add(instance)
        try:
            db.session.commit()
        except Exception as ex:
            logger.critical(f"Exception while committing changes to db: {ex}")
            db.session.rollback()
        return instance

    @staticmethod
    def encrypt(string):
        key = os.getenv('SECRET_KEY', default='secret!key')
        string = repr(string)  # converting all kind of object to string
        encoded_chars = []
        for i in range(len(string)):
            key_c = key[i % len(key)]
            encoded_c = chr(ord(string[i]) + ord(key_c) % 256)
            encoded_chars.append(encoded_c)
        encoded_string = "".join(encoded_chars)
        return base64.urlsafe_b64encode(encoded_string.encode()).decode()

    @staticmethod
    def decrypt(string):
        key = os.getenv('SECRET_KEY', default='secret!key')
        decoded_string = base64.urlsafe_b64decode(string).decode()
        decoded_chars = []
        for i in range(len(decoded_string)):
            key_c = key[i % len(key)]
            decoded_c = chr(ord(decoded_string[i]) - ord(key_c) % 256)
            decoded_chars.append(decoded_c)
        decoded_string = "".join(decoded_chars)
        return literal_eval(decoded_string)  # converting all string to their original state

    @classmethod
    def getSearchString(cls, user: User, searchString):
        if searchString is None:
            return

        # gathering all company ids of which user is part of. They are used to filter search.
        companies = [relation.company_id for relation in db.session.query(UserToCompany).filter_by(user_id=user.id).all()]
        try:  # if searchString is int than most probably it can be antragsnummer
            number = int(searchString)
            instance = db.session.query(cls).filter_by(
                antragsnummer=number).filter(cls.company_id.in_(companies)).order_by(cls.timestamp.desc()).first()
            if instance:
                return [instance]
        except:
            pass
        instances = {}

        # looping through all records of the companies in which current user is a part
        from sqlalchemy import func
        matching_instances = db.session.query(cls).filter(and_(cls.company_id.in_(companies), func.lower(
            cls.searchString).contains(func.lower(str(searchString))))).order_by(cls.timestamp.desc()).all()
        for obj in matching_instances:
            print(f'*** Found Antrags: {obj}')
            if not obj.antrag_id in instances:
                instances[obj.antrag_id] = obj
            else:
                if instances[obj.antrag_id].timestamp < obj.timestamp:  # if current object is latest than previously
                    instances[obj.antrag_id] = obj                      # stored then replace it with current object
        return list(instances.values())

    @classmethod
    def getInstancesFromAntragsnummer(cls, user, antragsnummer):
        companies = (relation.company_id for relation in
                     db.session.query(UserToCompany).filter_by(user_id=user.id).all())
        return db.session.query(cls).filter_by(
            antragsnummer=antragsnummer).filter(cls.company_id.in_(companies)).order_by(cls.timestamp.desc()).all()

    @staticmethod
    def getLatest(antrag_id):
        instance = db.session.query(AntragActivityRecords).filter_by(antrag_id=antrag_id).order_by(
            AntragActivityRecords.timestamp.desc()).first()
        if not instance:  # if no result from antrag_id then it might be record id. This is only for flexibility
            instance = db.session.query(AntragActivityRecords).filter_by(id=antrag_id).order_by(
                AntragActivityRecords.timestamp.desc()).first()
        db.session.expunge(instance)  # detaching instance from session so changes in it won't conflict with session
        instance.json_data = json.loads(instance.json_data)
        instance.json_data_activities = json.loads(instance.json_data_activities)
        decrypted_data = dict()
        for key, value in instance.json_data_activities.items():
            if value.get("encrypt"):  # if encrypt = True then this activity needs to be decrypted
                decrypted_data[key] = {}
                for ke, val in value.items():
                    if ke == "encrypt":
                        decrypted_data[key][ke] = val
                        continue
                    decrypted_data[key][AntragActivityRecords.decrypt(ke)] = AntragActivityRecords.decrypt(val)
            else:
                decrypted_data[key] = value
        instance.json_data_activities = decrypted_data  # replaced instance's dict with decrypted one
        return instance

    @classmethod
    def getAllVersions(cls, antragsnummer):
        number = int(antragsnummer)
        return db.session.query(cls).filter_by(antragsnummer=number).order_by(AntragActivityRecords.timestamp).all()

    def get_label(self):
        return ' '.join((
            self.searchString,
            self.antragsnummer,
            self.timestamp.strftime("%d.%m.%Y, %H:%M:%S"),
        ))

    def to_dict(self):
        dic = {
            "id": self.id,
            "user_id": self.user_id,
            "company_id": self.company_id,
            "antragsnummer": self.antragsnummer,
            "status": self.status,
            "timestamp": self.timestamp.strftime("%d.%m.%Y, %H:%M:%S"),
            "searchString": self.searchString,
            "json_data": json.loads(self.json_data)
        }
        return dic

    def to_json(self):
        return json.dumps(self.to_dict())


#
# Gamification Models
#
class GamificationEvent(db.Model):
    __tablename__ = 'gamification_events'
    id = db.Column(db.Integer, primary_key=True)
    name = db.Column(db.String(32), unique=True, nullable=False)

    def __str__(self):
        return self.name


class GamificationActivity(db.Model):
    __tablename__ = 'gamification_activities'
    id = db.Column(db.Integer, primary_key=True)
    created = db.Column(db.DateTime, nullable=False, default=datetime.utcnow)
    processed = db.Column(db.DateTime, nullable=True)
    is_processed = db.Column(db.Boolean, nullable=False, default=False)
    event_details = db.Column(db.String(1024), nullable=True)
    event_id = db.Column(db.Integer, db.ForeignKey('gamification_events.id'), nullable=False)
    user_id = db.Column(db.String(56), db.ForeignKey('users.id'), nullable=False)
    company_id = db.Column(db.String(56), db.ForeignKey('companies.id'), nullable=False)

    # relationships
    event = db.relationship('GamificationEvent', foreign_keys=[event_id])
    user = db.relationship('User', backref='gamification_activities', foreign_keys=[user_id])
    company = db.relationship('Company', backref='gamification_activities', foreign_keys=[company_id])

    def __str__(self):
        msg = f'{self.user.email}: {self.event.name}'
        if self.processed:
            return f'{msg} - processed at {self.processed.strftime("%d.%m.%Y, %H:%M:%S")}'
        else:
            return f'{msg} - NOT processed'

    @classmethod
    def new(cls, user, event, event_details):
        #
        # create instance based on the provided data
        #

        if not isinstance(event, GamificationEvent):
            if isinstance(event, int):
                event = db.session.query(GamificationEvent).filter_by(id=event).first()
            elif isinstance(event, str):
                event = db.session.query(GamificationEvent).filter_by(name=event).first()

        instance = cls(
            user_id=user.id,
            company_id=user.company_id,
            event=event,
            event_details=event_details,
        )

        try:
            db.session.add(instance)
            db.session.commit()
        except Exception as ex:
            logger.error(f"Error during Commit: {ex}. Rolling back.")
            db.session.rollback()

        return instance

    def set_processed(self, commit=True):
        #
        # set activity as processed
        #

        self.processed = datetime.utcnow()
        self.is_processed = True

        # save to db
        if commit:
            try:
                db.session.commit()
            except Exception as ex:
                logger.critical(f"Exception while committing changes to db: {ex}")
                db.session.rollback()


class GamificationUserStats(db.Model):
    __tablename__ = 'gamification_statistics'
    user_id = db.Column(db.String(56), db.ForeignKey('users.id'), primary_key=True, nullable=False)
    company_id = db.Column(db.String(56), db.ForeignKey('companies.id'), primary_key=True, nullable=False)
    type_id = db.Column(db.Integer, db.ForeignKey('gamification_badge_types.id'), primary_key=True)
    daily = db.Column(db.Integer, default=0)
    weekly = db.Column(db.Integer, default=0)
    monthly = db.Column(db.Integer, default=0)
    yearly = db.Column(db.Integer, default=0)
    all_time = db.Column(db.Integer, default=0)
    last_updated = db.Column(db.DateTime, default=datetime.now, onupdate=datetime.now)

    user = db.relationship('User', backref='gamification_statistics', foreign_keys=[user_id])
    company = db.relationship('Company', backref='gamification_statistics', foreign_keys=[company_id])
    type = db.relationship('GamificationBadgeType', backref="gamification_badge_types", foreign_keys=[type_id])

    def __str__(self):
        return self.user_id + f" - {self.type_id} -" + " -- " + str(self.get_user_statistics())

    def check_timeline(self):
        year_is_changed = self.last_updated.date().year < datetime.now().date().year
        # year_is_changed used when year changed but month or week number is same
        changes = False
        if self.last_updated.date() < datetime.now().date() or year_is_changed:
            self.daily = 0
            changes = True
        if self.last_updated.date().isocalendar()[1] < datetime.now().date().isocalendar()[1] or year_is_changed:
            self.weekly = 0
            changes = True
        if self.last_updated.date().month < datetime.now().date().month or year_is_changed:
            self.monthly = 0
            changes = True
        if year_is_changed:
            self.yearly = 0
            changes = True
        if changes:  # only update last_update time when it is needed
            self.last_updated = datetime.now()
            try:
                db.session.commit()
            except Exception as ex:
                logger.critical(f"Exception while committing changes to db: {ex}")
                db.session.rollback()

    def add_points(self, points=1, weight=1):
        self.check_timeline()
        weighted_point = points * weight
        self.daily += weighted_point
        self.weekly += weighted_point
        self.monthly += weighted_point
        self.yearly += weighted_point
        self.all_time += weighted_point
        self.update_badge()
        return {"all time": self.all_time}

    def get_user_statistics(self):
        self.check_timeline()
        json_data = {"daily": self.daily, "weekly": self.weekly, "monthly": self.monthly,
                     "annual": self.yearly, "all time": self.all_time}
        return json_data

    @classmethod
    def new(cls, user_id, company_id, type_id):
        instance = cls(
            user_id=user_id,
            company_id=company_id,
            type_id=type_id,
        )
        db.session.add(instance)
        try:
            db.session.commit()
        except Exception as ex:
            logger.critical(f"Exception while committing changes to db: {ex}")
            db.session.rollback()

        return instance

    def update_badge(self):
        level_id = self.get_level_id()
        badge = db.session.query(GamificationBadge).filter_by(user_id=self.user_id).filter_by(
            company_id=self.company_id).filter_by(type_id=self.type_id).filter_by(level_id=level_id).first()
        if not badge:
            badge = GamificationBadge(
                user_id=self.user_id, company_id=self.company_id, type_id=self.type_id, level_id=level_id)
            db.session.add(badge)
            try:
                db.session.commit()
            except Exception as ex:
                logger.critical(f"Exception while committing changes to db: {ex}")
                db.session.rollback()
            ToastNotifications.new(
                message=badge.id, type="badge", company_ids=self.company_id, user_ids=self.user_id
            )

    def get_level_id(self):
        level = db.session.query(GamificationBadgeLevel).filter(and_(self.all_time >= GamificationBadgeLevel.min_level,
            (or_(self.all_time <= GamificationBadgeLevel.max_level, GamificationBadgeLevel.max_level == None)))).first()
        return level.id

    @staticmethod
    def get_type_id(event: GamificationEvent, event_details):
        eventName = event.name.lower()
        event_details = json.loads(event_details)
        eventType = event_details.get("lineOfBusiness")
        if "policy" in eventName and eventType:
            return db.session.query(GamificationBadgeType).filter_by(name=f"Polizze {eventType}").first().id
        elif "antrag" in eventName and eventType:
            return db.session.query(GamificationBadgeType).filter_by(name=f"Antrag {eventType}").first().id
        elif "login" in eventName:
            return db.session.query(GamificationBadgeType).filter_by(name="Login").first().id
        elif "policy" in eventName:
            return db.session.query(GamificationBadgeType).filter_by(name=f"Policy").first().id
        elif "antrag" in eventName:
            return db.session.query(GamificationBadgeType).filter_by(name=f"Antrag").first().id
        else:
            print(f"{eventName} event is ignored.")
            return None

    @staticmethod
    def get_weight(event: GamificationEvent, event_details):
        eventName = event.name
        event_details = json.loads(event_details)
        lob = event_details.get("lineOfBusiness", "")
        activityName = event_details.get("Activity")
        if lob and activityName:
            id_ = db.session.query(GamificationActivityWeight).filter(and_(
                GamificationActivityWeight.activity_name == activityName,
                GamificationActivityWeight.line_of_business == lob)).first()
        else:
            if lob and eventName:
                id_ = db.session.query(GamificationActivityWeight).filter(and_(
                    GamificationActivityWeight.activity_name == eventName,
                    GamificationActivityWeight.line_of_business == lob)).first()
            else:
                id_ = db.session.query(GamificationActivityWeight).filter_by(activity_name=eventName).first()
        if not id_:
            print(f"Combination of activityName: {str(activityName)}, lineOfBusiness: {str(lob)} or "
                  f"event: {str(eventName)} not found in Weight table. Using default value 10.")
            id_ = db.session.query(GamificationActivityWeight).filter_by(activity_name="default").first()
            if not id_:
                print("Default weight not found in database. Returning 10")
                return 10
        return id_.points

    @classmethod
    def create_or_update_row(cls, activity: GamificationActivity, commit=True):
        type_id = cls.get_type_id(activity.event, activity.event_details)
        weightage = cls.get_weight(activity.event, activity.event_details)
        if not type_id:
            return None
        user_id = activity.user_id
        company_id = activity.company_id
        user_stats = db.session.query(GamificationUserStats).filter_by(user_id=user_id
            ).filter_by(company_id=company_id).filter_by(type_id=type_id).first()
        if user_stats:
            user_stats.add_points(weight=weightage)
        if not user_stats:
            user_stats = GamificationUserStats.new(user_id=user_id, company_id=company_id, type_id=type_id)
            user_stats.add_points(weight=weightage)
        if commit:
            try:
                db.session.commit()
            except Exception as ex:
                logger.critical(f"Exception while committing changes to db: {ex}")
                db.session.rollback()
        return user_id

    @staticmethod
    def commit():
        try:
            db.session.commit()
        except Exception as ex:
            logger.critical(f"Exception while committing changes to db: {ex}")
            db.session.rollback()

## Badges

class GamificationBadgeDescription(db.Model):
    #
    # Badge earn requirements for given Type and Level
    #
    __tablename__ = 'gamification_badge_descriptions'
    type_id = db.Column(db.Integer, db.ForeignKey('gamification_badge_types.id'), primary_key=True)
    level_id = db.Column(db.Integer, db.ForeignKey('gamification_badge_levels.id'), primary_key=True)
    description = db.Column(db.String(512), nullable=False)

    # relationships
    type = db.relationship('GamificationBadgeType', backref='descriptions', foreign_keys=[type_id])
    level = db.relationship('GamificationBadgeLevel', foreign_keys=[level_id])

class GamificationBadgeLevel(db.Model):
    __tablename__ = 'gamification_badge_levels'
    id = db.Column(db.Integer, primary_key=True)
    name = db.Column(db.String(32), unique=True, nullable=False)
    min_level = db.Column(db.Integer, unique=True, nullable=False)
    max_level = db.Column(db.Integer, unique=True, nullable=True)
    is_lowest = db.Column(db.Boolean, nullable=False, default=False,)
    next_level_id = db.Column(db.Integer, db.ForeignKey('gamification_badge_levels.id'), nullable=True)

    #relationships
    next_level = db.relationship('GamificationBadgeLevel', remote_side=[next_level_id], uselist=False)

    def __str__(self):
        return self.name

    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.max_level = self.get_next_level_min()

    def get_next_level_name(self):
        if self.next_level:
            return self.next_level.name

    def get_next_level_min(self):
        if self.next_level:
            return self.next_level.min_level - 1
        return None


class GamificationBadgeType(db.Model):
    __tablename__ = 'gamification_badge_types'
    id = db.Column(db.Integer, primary_key=True)
    name = db.Column(db.String(32), unique=True, nullable=False)
    title = db.Column(db.String(32), nullable=False)

    # reletionaships
    #descriptions = db.relationship(
    #    'GamificationBadgeDescription',
    #    primaryjoin="GamificationBadgeType.id==GamificationBadgeDescription.type_id",
    #)

    def get_description(self):
        return {(d.level.name if not d.level.is_lowest else 'lowest'): d.description for d in self.descriptions}

    def __str__(self):
        return self.name

    def to_json(self):
        return {
            'name': self.name,
            'title': self.title,
            'description': self.get_description(),
        }

class GamificationBadge(db.Model):
    __tablename__ = 'gamification_badges'
    id = db.Column(db.Integer, primary_key=True)
    user_id = db.Column(db.String(56), db.ForeignKey('users.id'), nullable=False)
    company_id = db.Column(db.String(56), db.ForeignKey('companies.id'), nullable=False)
    type_id = db.Column(db.Integer, db.ForeignKey('gamification_badge_types.id'), nullable=False)
    level_id = db.Column(db.Integer, db.ForeignKey('gamification_badge_levels.id'), nullable=False)
    achieved_at = db.Column(db.DateTime, nullable=False, default=datetime.utcnow)
    is_seen = db.Column(db.Boolean, nullable=False, default=False)

    # relationships
    user = db.relationship('User', backref='gamification_badges', foreign_keys=[user_id])
    company = db.relationship('Company', foreign_keys=[company_id])
    type = db.relationship('GamificationBadgeType', foreign_keys=[type_id])
    level = db.relationship('GamificationBadgeLevel', foreign_keys=[level_id])

    def __str__(self):
        return f'Badge {self.level.name} {self.type.name} for {self.user.email}'

    def set_seen(self):
        self.is_seen = True
        try:
            db.session.commit()
        except Exception as ex:
            logger.critical(f"Exception while committing changes to db: {ex}")
            db.session.rollback()

    def to_json(self):
        return {
            'type': self.type.name,
            'level': self.level.name,
            'next_level': self.level.get_next_level_name(),
            'isSeen': self.is_seen,
        }

    @classmethod
    def get_user_stats(cls, user_id=user_id):
        user = db.session.query(cls).filter_by(user_id=user_id).all()
        json_data = [data.to_json() for data in user]
        return json_data


class GamificationActivityWeight(db.Model):
    __table_name__ = "gamification_activity_weight"
    __table_args__ = (db.UniqueConstraint('activity_name', 'line_of_business', name='unique'),)
    id = db.Column(db.Integer, primary_key=True)
    activity_name = db.Column(db.String(32), nullable=False)
    line_of_business = db.Column(db.String(16), nullable=True)
    points = db.Column(db.Integer, nullable=False)

    @classmethod
    def get_points(cls, id):
        if id:
            weight = cls.query.filter(cls.id == id).first()
            if weight:
                return weight.points
        print("Weightage id is either not supplied or not found in Weightage table. Returned default value of 1")
        return 1

    @classmethod
    def new(cls, activity_name, line_of_business, points=1):
        instance = cls(activity_name=activity_name, line_of_business=line_of_business, points=points)
        db.session.add(instance)
        try:
            db.session.commit()
        except Exception as ex:
            logger.critical(f"Exception while committing changes to db: {ex}")
            db.session.rollback()

    def to_json(self):
        js = {
            "id": self.id,
            "activity_name": self.activity_name,
            "line_of_business": self.line_of_business,
            "points": self.points
        }
        return js

    @classmethod
    def get_all_id_points(cls):
        all_rows = cls.query.all()
        all_data = {row.id: row.points for row in all_rows}
        return all_data


class ToastNotifications(db.Model):
    __tablename__ = "toast_notifications"
    id = db.Column(db.String(56), primary_key=True, default=generate_id)
    company_id = db.Column(db.String(56), db.ForeignKey('companies.id'), primary_key=True)
    user_id = db.Column(db.String(56), db.ForeignKey('users.id'), primary_key=True)
    message = db.Column(db.String(128), nullable=False)
    type = db.Column(db.String(16), default="default", nullable=False, server_default="default")
    duration = db.Column(db.Integer, nullable=True)
    horizontal = db.Column(db.String(16), default="left", nullable=False, server_default="left")
    vertical = db.Column(db.String(16), default="top", nullable=False, server_default="top")
    created = db.Column(db.DateTime, nullable=False, default=datetime.utcnow)
    seen_at = db.Column(db.DateTime)

    @classmethod
    def new(cls, message, type="default", duration=3000, horizontal="left", vertical="top", company_ids=None, user_ids=None):
        companies = []
        if company_ids:
            if isinstance(company_ids, list):
                for company_id in company_ids:
                    if db.session.query(Company).filter_by(id=company_id).first():
                        companies.append(company_id)
                    else:
                        print(f"Company id {company_id} is not correct. Please recheck.")
            else:
                if db.session.query(Company).filter_by(id=company_ids).first():
                    companies.append(company_ids)
                else:
                    print(f"Company id {company_ids} is not correct. Please recheck.")
        else:
            companies = [company.id for company in db.session.query(Company).all()]
        if not companies:
            print("Supplied company ids are not valid. Hence no notification will be updated")
        for company_id in companies:
            users = []
            if not user_ids:
                users = [user.id for user in db.session.query(User).filter_by(company_id=company_id).all()]
            else:
                if isinstance(user_ids, list):
                    for user_id in user_ids:
                        if db.session.query(User).filter_by(id=user_id).first():
                            users.append(user_id)
                        else:
                            print(f"User id {user_id} is not correct. Please recheck.")
                else:
                    if db.session.query(User).filter_by(id=user_ids).first():
                        users.append(user_ids)
                    else:
                        print(f"User id {user_ids} is not correct. Please recheck.")
            for user_id in users:
                if type == "badge":
                    instance = cls(company_id=company_id, user_id=user_id, message=message, type=type)
                else:
                    instance = cls(company_id=company_id, user_id=user_id, message=message, type=type,
                                   duration=duration, horizontal=horizontal, vertical=vertical)
                db.session.add(instance)
        try:
            db.session.commit()
        except Exception as ex:
            logger.critical(f"Exception while committing changes to db: {ex}")
            db.session.rollback()

    def set_seen(self):
        self.seen_at = datetime.now()
        try:
            db.session.commit()
        except Exception as ex:
            logger.critical(f"Exception while committing changes to db: {ex}")
            db.session.rollback()


class AntragNummer(db.Model):
    count = db.Column(db.Integer, default=10000000, primary_key=True)

    @staticmethod
    def get_count():
        count = db.session.query(AntragNummer).first().count
        AntragNummer.update_count()  # updating 100 as it calls count, because of program stops before completing 100,
        return count                 # then current base value will be resent and it will create same ids

    @staticmethod
    def update_count():
        instance = db.session.query(AntragNummer).first()
        instance.count += 100
        db.session.add(instance)
        try:
            db.session.commit()
        except Exception as ex:
            logger.critical(f"Exception while committing changes to db: {ex}")
            db.session.rollback()



from polzybackend.models import db, GamificationUserStats, GamificationActivityWeight
from polzyFunctions.GlobalConstants import logger


class Singleton(type):
    _instances = {}

    def __call__(cls, *args, **kwargs):
        if cls not in cls._instances:
            cls._instances[cls] = super(Singleton, cls).__call__(*args, **kwargs)
        return cls._instances[cls]


class HitList(metaclass=Singleton):
    def __init__(self):
        pass

    def deriveUserRanking(self, user):
        #
        # generate ranking for user
        #

        rank_categories = [
            'daily',
            'weekly',
            'monthly',
            'annual',
            'all time'
        ]

        # Get all gamification statistics for same company
        stats = db.session.query(GamificationUserStats).filter_by(company_id=user.company_id).all()
        ranking_dic = {}

        for rank_category in rank_categories:  # creating response json with rank categories as key
            ranking_dic[rank_category] = self.get_rank_by_category(rank_category, stats)

        # generate and return random ranking
        return ranking_dic

    def get_rank_by_category(self, rank_category, stats):
        # getting weighted points for every activity and adding them up per user
        unadded_user_points = {}
        for stat in stats:  # first getting weighted points of each activity
            try:
                if not stat.user.email in unadded_user_points:
                    unadded_user_points[stat.user.email] = []
            except AttributeError:
                logger.critical(f"User id {stat.user_id} doesn't exist. So ranking is skipping it.")
                continue
            unadded_user_points[stat.user.email].append([stat.get_user_statistics().get(
                rank_category, 0), stat.user.email])

        added_points = []
        for user_email in unadded_user_points:  # adding activity points to one per user
            score = 0
            for points in unadded_user_points[user_email]:
                score += points[0]
            added_points.append([score, user_email])

        last_score = -1
        last_rank = 0
        ranked_data = []
        for dt in sorted(added_points, reverse=True):  # ranking user as per their points
            dic = {}
            if dt[0] != last_score:  # give same rank number if last iteration has same score
                last_rank += 1
                last_score = dt[0]
                dt.append(last_rank)
            else:
                dt.append(last_rank)
            dic['name'] = dt[1]
            dic['operations'] = dt[0]
            dic['rank'] = dt[2]
            ranked_data.append(dic)
        return ranked_data



from dataclasses import dataclass


@dataclass
class AntragsStatus:
    """
    The central handling of status of a fast offer. We start with "new", etc.
    Can be subclassed in each application.

    The status changes usually happen in method execute_activity of activities, when successful.

    """
    statusDict = {
        000: "New",
        100: "Calculated",
        150: "Printed",
        400: "Customer registered",
        450: "Application questions answered",
        500: "Documents printed",
        550: "Documents signed",
        800: "Sent"
    }
    s000_new = "New"
    s100_calculated = "Calculated"
    s150_printed = "Printed"
    s400_RegisteredCustomer = "Customer registered"
    s450_QuestionsAnswered = "Application questions answered"
    s500_DocumentsPrinted = "Documents printed"
    s550_DocumentsSigned = "Documents signed"
    s800_sent = "Sent"

    @classmethod
    def getStatusNumberFromText(cls, statusText) -> int:
        """
        Used to get the numeric representation of a status text.

        :param statusText: Text of the status, e.g. "New"
        :return: Numeric value of the status, e.g. 0
        """
        for key, value in cls.statusDict.items():
            if value == statusText:
                return key




import threading

lock = threading.Lock()  # to make class thread safe


class Singleton(type):
    _instances = {}

    def __call__(cls, *args, **kwargs):
        if cls not in cls._instances:
            with lock:
                if cls not in cls._instances:
                    cls._instances[cls] = super(Singleton, cls).__call__(*args, **kwargs)
        return cls._instances[cls]


class Counter(metaclass=Singleton):
    """
    This is a singleton class to provide antragnummer. It takes care of providing unique number/id.
    """
    def __init__(self):
        from polzybackend.models import AntragNummer  # importing table here to avoid import error
        self.AntragNummer = AntragNummer              # To avoid importing this table on every method call we made it
        self.number = self.AntragNummer.get_count()                                                 # class attribute

    def get_number(self):
        number = self.number
        self.number += 1
        if self.number % 100 == 0:            # if number is divisible by 100 update value in db with 100
            self.AntragNummer.update_count()
        return number



from polzyFunctions.GlobalConstants import logger
from polzyFunctions.Dataclasses.Antrag import Antrag
# Also import all the possible activites candidate here, they will be dynamically imported


class AntragActivitiesDerive:
    """
    Determines which activities are possible based on the application status, product, etc.

    If there are already activities for this application, then use the existing activities accordingly,
    do not re-instantiate.

    This class is an example and you should create a sub class to achieve goal by referring to this implementation.
    """
    def __init__(self, antrag: Antrag):
        self.antrag = antrag
        self.possibleActivities = []  # store possible activities in this

        existingActivities = []  # store already existing activities in this. It will be used to avoid re-instantiate
        if hasattr(antrag, "Activities"):
            for activity in antrag.Activities:  # loop through already existing activity and store their class name
                existingActivities.append(activity.__class__.__name__)

        possibleActivitiesCandidates = []  # ["activity1", "activity2"] store possible activities name in string here

        alreadyInitiatedActivities = []  # store activities which are not possible with current status here

        if hasattr(antrag, "Activities"):
            for activity in possibleActivitiesCandidates:
                existingActivity = self.antrag.get_activity(activity, optional=True)
                if existingActivity:
                    # This activity was already there, do not re-instantiate.
                    alreadyInitiatedActivities.append(activity)

            for alreadyInitiatedActivity in alreadyInitiatedActivities:
                possibleActivitiesCandidates.remove(alreadyInitiatedActivity)

        # In PossibleActivities there are now all activities that are theoretically possible, but not yet
        # were instantiated. these must now be generated.
        for newActivity in possibleActivitiesCandidates:
            tmpClassReference = globals()[newActivity]
            self.possibleActivities.append(tmpClassReference(antrag=antrag))  # initiate and store activity

        # In getActivitiesforAntrag we return ALL activities for the application, so we have to
        # Now add the existing activities again.
        if hasattr(antrag, "Activities"):
            antrag.Activities.extend(self.possibleActivities)
            self.possibleActivities = antrag.Activities

    def getActivitiesForAntrag(self):
        """
        Go through all the activities and ask them if they want to be active
        :return:
        """
        # call method checkIsActivityValidForStatus of activity to know if it is suitable for current stats
        self.possibleActivities = [x for x in self.possibleActivities if x.checkIsActivityValidForStatus()]

        logger.debug(f"List of possible Activities for this Antrag in state {self.antrag.status} is: \n{self.possibleActivities}")
        return self.possibleActivities



from polzybackend.mediators import Antrag
from polzyFunctions.Dataclasses.Antrag import Antrag as AntragBase
from polzyFunctions.Activities import AntragActivity
from polzyFunctions.Activities.Activity import Activity
from polzyFunctions.GlobalConstants import GlobalConstants, logger
from flask import current_app, url_for


class VIGAntrag(Antrag):
    #
    # local implementation of PoLZy antrag class
    #

    def initialize(self):
        #
        # initializes antrag instance
        #
        # get antrag object based on product_name

        self.instance = AntragBase(self.user)
        self.id = str(self.instance.id)

    def load(self, client):  # client is supplied as routes can supply it to some inherited version. So to avoid error
        try:
            self.instance = AntragBase(self.user)
        except Exception as e:
            current_app.logger.exception(f"Load of antrag instance failed: {e}")
            raise ValueError("Load of antrag instance failed")

    def __str__(self):
        return f'<Antrag {self.id}: {self.product_name}>'

    def get(self):
        #
        # returns antrag instance as json object to front-end 
        #

        result = self.instance.parseToFrontend()
        return result

    def clone(self):
        #
        # returns a deepcopy of the current instance with updated props
        #

        antrag_copy = super().clone()
        antrag_copy.instance.updateAfterClone()
        antrag_copy.id = str(antrag_copy.instance.id)

        return antrag_copy

    def updateFields(self, data):
        #
        # updates antrag fields based on data
        # If data.get('activity') is None, than the data of the Antragsinstance needs to be updated with new
        # field values.
        #
        # If data.get("activity") is set, then we need to update into an activity.
        # 16.3.21: This works well.

        if data.get('activity') is None:
            self.instance.updateFieldValues(data.get('values'))
        else:
            # get activity class
            activity_class = AntragActivity
            if activity_class is None:
                raise ValueError(f'Unknown activity: {data["activity"]}')

            # update activity values for the activity-instance inside Antrag
            activity = self.instance.get_activity(activity_class.__name__)
            if activity:
                activity[0].updateFieldValues(data.get('values'))

    def findActivity(self, activity_name):
        try:
            target_activity = AntragActivity
            lActivity = next(filter(lambda x: isinstance(x, target_activity), self.instance.Activities), None)
        except TypeError:
            logger.exception(f"Didn't find anything useful for activity {activity_name}. If it's a new activity, then"
                             f" make sure you imported it in antrag.py and added it to activity_classes on the top. ")
            return None

        logger.debug(f"Found existing activity inside Antrag: {lActivity.__class__.__name__}")
        return lActivity

    def executeActivity(self, data):
        #
        # executes antrag activity defined in data
        #
        # update values of Antragsinstance for Frontend display after the activity was executed.
        #
        # We have a long list of IF-conditions because many activities have different return parameters.

        current_activity = self.findActivity(data['activity'])
        if not current_activity:
            print(f"Activity couldn't be found. Something weird going on. data from Frontend: {data}")
            current_app.logger.critical(f"Activity couldn't be found. Something weird going on. data from Frontend: {data}")

        if data['activity'] == 'Berechnen':               # calculate
            # In Berechnen we get the data for the antrag, so we must provide it to Antags-Instance
            self.instance.updateFieldValues(data.get('values'))
            current_activity.executeActivity()
            self.instance.fillCurrentlyPossibleActivities()
            return self.get()

        current_activity.updateFieldValues(data.get('values'))
        lResult = current_activity.executeActivity()

        if data["activity"] in ["Eurotax Vollsuche", "Empfehlung!", "Dokumente", 'Antrag erzeugen PM+']:
            self.instance.fillCurrentlyPossibleActivities()
            return self.get()

        if data['activity'] == 'Drucken':               # print
            if not lResult:
                raise ValueError('Unable to create PDF')
            return {
                "link": url_for(
                    'support.downloads',
                    filename=current_activity.getFolderAndPath(),
                    _external=True,
                ),
            }
        elif data['activity'] == 'Antrag erzeugen':       # save to VNG
            return {
                "link": current_activity.getLinkOfRealApplication()
            }

        elif data['activity'] == "Deckungsuebersicht":
            return {
                "link": url_for(
                    'support.downloads',
                    filename=current_activity.getFolderAndPath(),
                    _external=True,
                ),
            }

        elif data['activity'] == "VN festlegen":
            # Update VN-Fields inside the antrags instance with the latest values.
            self.instance.updateFieldValues(data.get('values'))
            current_activity.setStatusOfAntragAfterPartnerAssignment()

            self.instance.fillCurrentlyPossibleActivities()
            return self.get()

        elif data["activity"] in ["Partner festlegen", "Antragsfragen BUFT"]:
            if lResult:
                self.instance.fillCurrentlyPossibleActivities()
            else:
                # Execution of activity was not successful, activity should stay open
                current_activity.postExecuteBehavior = Activity.POSTEXECUTION_ACTIVE
            return self.get()

        else:
            logger.exception(f'Activity was {data.get("activity")}. No logic implemented for that')
            raise ValueError(f'Antrag activiy {data.get("activity")} failed')

    def getValueList(self, listName):
        if listName == "firmenArten":
            return sorted(["GmbH", "GmbH&Co KG", "AG", "ArGE", "Einz.Unt", "EWIV", "Gen", "GenbR", "E.U", "KEG",
                           "KG", "OEG", "OG", "OHG", "OPG", "PsT", "Sparkasse", "Verein", "VVaG", "WEG"])
            
        lReasult = self.instance.getValueList(listName)
        return lReasult

    def getRemoteDocuments(self, documents_id: list):
        #
        # returns path to fetched remote document 
        #

        path_to_result_document = self.instance.getRemoteDocuments(documents_id)
        return path_to_result_document

    def generateEml(self, documents_id: list):
        # generate eml file with documents as attachments
        path_to_eml = self.instance.generateEml(documents_id)
        return path_to_eml

    def generateAntragEml(self):
        # generate eml file with antrag link
        return self.instance.generateAntragEml()



from polzyFunctions.utils import get_file_path, Singleton
from polzyFunctions.GlobalConstants import logger
from polzyFunctions.translator.DataHandler import Data


class Translator(metaclass=Singleton):
    """
    We are making this class singleton to share same default language on multiple instances.
    """
    def __init__(self, default="en"):
        logger.debug("Language translation backend module initialized - buffered")
        self.default_language = default.lower()
        self.data = Data().data  # getting data from singleton class's attribute

    def translate(self, word, language=None):
        if not language:
            language = self.default_language
        language = language.lower()
        if language == "en":
            return word
        result = self.data.get(word).get(language)
        if not result:
            result = word
            logger.info(f'Translation of "{word}" for language "{language}" not found!')
        return result

    def add_translation_dict(self, dict):
        # this method is to be used when we need to update translation from dictionary. It can be called in sub repo
        # for customer specific translations.
        # structure of input dictionary: {"English word": {"de": "translation for it", "wi": "translation for it", ...}}
        self.data.update(dict)

    def add_translation_file(self, fileNameAndPath):
        # this method is to be used when we need to update translation from json file. It can be called in sub repo
        # for customer specific translations.
        # structure in input json file: {"English word": {"de": "translation for it", "wi": "translation for it", ...}}
        fileNameAndPath = get_file_path(fileNameAndPath)
        dic = Data.get_data(fileNameAndPath)
        print(f"Updating translation data of file {fileNameAndPath}, data: {dic}")
        self.add_translation_dict(dic)

    def update_default_language(self, language):
        # this method is planned to be used to update default language of translation module as per current user
        self.default_language = language.lower()
